from typing import Any, Callable, ClassVar, Final

import numpy as np
import torch

import lucid
from lucid.test.core import TensorOpWithTorchBase, TensorOpTorchCase

_UNARY_FLOAT_2x2: Final[list[list[float]]] = [[1.0, 2.0], [3.0, 4.0]]
_UNARY_INT_2x2: Final[list[list[int]]] = [[1, 2], [3, 4]]
_UNARY_NEG_2x2: Final[list[list[float]]] = [[-1.0, -2.0], [3.0, -4.0]]

_UNARY_ARC_RANGE: Final[list[list[float]]] = [[-0.75, -0.25], [0.25, 0.75]]
_UNARY_LOG_RANGE: Final[list[list[float]]] = [[1.0, 2.0], [3.0, 4.0]]
_UNARY_CUM_2x3: Final[list[list[float]]] = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]

_UNARY_RECT_2x3: Final[list[list[float]]] = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]
_UNARY_DECIMALS: Final[list[list[float]]] = [[1.14, 2.73], [3.36, 4.99]]
_UNARY_CLIP_1B: Final[list[list[float]]] = [[-1.0, 0.5], [1.5, 2.5]]


class _TestUnaryFuncOpBase(TensorOpWithTorchBase):
    _collect_nested_test_classes = True
    case_name: ClassVar[str]
    input: ClassVar[Any]

    forward_op: ClassVar[Callable[[Any], Any]]
    torch_forward: ClassVar[Callable[[Any], Any]]
    expected_forward: ClassVar[Any]

    requires_grad: ClassVar[bool] = True
    dtype: ClassVar[Any | None] = None
    compare_grad_with_torch: ClassVar[bool] = True
    check_finite: ClassVar[bool] = True
    expected_input_grads: ClassVar[tuple[object | None, ...] | None] = None
    rtol: ClassVar[float] = 1e-7
    atol: ClassVar[float] = 1e-8

    def tensor_op_devices(self) -> tuple[str, ...]:
        return ("cpu", "gpu")

    def _build_unary_inputs(
        self,
        input_data: Any,
        *,
        dtype: Any | None = None,
        requires_grad: bool = True,
    ) -> tuple[Any]:
        if dtype is None:
            return (self.tensor(input_data, requires_grad=requires_grad),)
        return (
            self.tensor(
                input_data,
                dtype=dtype,
                requires_grad=requires_grad,
            ),
        )

    def _build_case(self) -> TensorOpTorchCase:
        forward_op = vars(type(self)).get("forward_op")
        torch_forward = vars(type(self)).get("torch_forward")

        if isinstance(forward_op, staticmethod):
            forward_op = forward_op.__func__
        if isinstance(torch_forward, staticmethod):
            torch_forward = torch_forward.__func__

        return TensorOpTorchCase(
            name=self.case_name,
            build_inputs=lambda _case: self._build_unary_inputs(
                self.input,
                dtype=self.dtype,
                requires_grad=self.requires_grad,
            ),
            forward_op=forward_op,
            expected_forward=self.expected_forward,
            expected_input_grads=self.expected_input_grads,
            requires_grad=(self.requires_grad,),
            check_finite=self.check_finite,
            torch_forward=torch_forward,
            compare_grad_with_torch=self.compare_grad_with_torch,
            rtol=self.rtol,
            atol=self.atol,
        )

    def tensor_op_cases(self) -> tuple[TensorOpTorchCase, ...]:
        return (self._build_case(),)


class _TestUnaryNoGradFuncOpBase(_TestUnaryFuncOpBase):
    requires_grad: ClassVar[bool] = False
    compare_grad_with_torch: ClassVar[bool] = False


class TestPow(_TestUnaryFuncOpBase):
    case_name = "pow"
    input = _UNARY_FLOAT_2x2
    forward_op = staticmethod(lambda a: a**2.0)
    torch_forward = staticmethod(lambda a: torch.pow(a, 2.0))
    expected_forward = np.array([[1.0, 4.0], [9.0, 16.0]], dtype=np.float64)


class TestRpow(_TestUnaryFuncOpBase):
    case_name = "rpow"
    input = _UNARY_FLOAT_2x2
    forward_op = staticmethod(lambda a: 2.0**a)
    torch_forward = staticmethod(lambda a: torch.pow(2.0, a))
    expected_forward = np.array([[2.0, 4.0], [8.0, 16.0]], dtype=np.float64)


class TestNeg(_TestUnaryFuncOpBase):
    case_name = "neg"
    input = _UNARY_FLOAT_2x2
    forward_op = staticmethod(lambda a: -a)
    torch_forward = staticmethod(lambda a: -a)
    expected_forward = np.array([[-1.0, -2.0], [-3.0, -4.0]], dtype=np.float64)


class TestInvert(_TestUnaryNoGradFuncOpBase):
    case_name = "invert_no_grad"
    input = _UNARY_INT_2x2
    forward_op = staticmethod(lambda a: ~a)
    torch_forward = staticmethod(torch.bitwise_not)
    dtype = lucid.Int64
    expected_forward = np.array([[-2, -3], [-4, -5]], dtype=np.int64)


class TestExp(_TestUnaryFuncOpBase):
    case_name = "exp"
    input = _UNARY_FLOAT_2x2
    forward_op = staticmethod(lucid.exp)
    torch_forward = staticmethod(torch.exp)
    expected_forward = np.array(np.exp(_UNARY_FLOAT_2x2), dtype=np.float64)


class TestLog(_TestUnaryFuncOpBase):
    case_name = "log"
    input = _UNARY_LOG_RANGE
    forward_op = staticmethod(lucid.log)
    torch_forward = staticmethod(torch.log)
    expected_forward = np.array(np.log(_UNARY_LOG_RANGE), dtype=np.float64)


class TestLog2(_TestUnaryFuncOpBase):
    case_name = "log2"
    input = _UNARY_LOG_RANGE
    forward_op = staticmethod(lucid.log2)
    torch_forward = staticmethod(torch.log2)
    expected_forward = np.array(np.log2(_UNARY_LOG_RANGE), dtype=np.float64)


class TestSqrt(_TestUnaryFuncOpBase):
    case_name = "sqrt"
    input = _UNARY_LOG_RANGE
    forward_op = staticmethod(lucid.sqrt)
    torch_forward = staticmethod(torch.sqrt)
    expected_forward = np.array(np.sqrt(_UNARY_LOG_RANGE), dtype=np.float64)


class TestSin(_TestUnaryFuncOpBase):
    case_name = "sin"
    input = _UNARY_FLOAT_2x2
    forward_op = staticmethod(lucid.sin)
    torch_forward = staticmethod(torch.sin)
    expected_forward = np.array(np.sin(_UNARY_FLOAT_2x2), dtype=np.float64)


class TestCos(_TestUnaryFuncOpBase):
    case_name = "cos"
    input = _UNARY_FLOAT_2x2
    forward_op = staticmethod(lucid.cos)
    torch_forward = staticmethod(torch.cos)
    expected_forward = np.array(np.cos(_UNARY_FLOAT_2x2), dtype=np.float64)


class TestTan(_TestUnaryFuncOpBase):
    case_name = "tan"
    input = _UNARY_FLOAT_2x2
    forward_op = staticmethod(lucid.tan)
    torch_forward = staticmethod(torch.tan)
    expected_forward = np.array(np.tan(_UNARY_FLOAT_2x2), dtype=np.float64)
    rtol = 1e-6
    atol = 1e-6


class TestArcsin(_TestUnaryFuncOpBase):
    case_name = "arcsin"
    input = _UNARY_ARC_RANGE
    forward_op = staticmethod(lucid.arcsin)
    torch_forward = staticmethod(torch.arcsin)
    expected_forward = np.array(np.arcsin(_UNARY_ARC_RANGE), dtype=np.float64)
    rtol = 1e-6
    atol = 1e-6


class TestArccos(_TestUnaryFuncOpBase):
    case_name = "arccos"
    input = _UNARY_ARC_RANGE
    forward_op = staticmethod(lucid.arccos)
    torch_forward = staticmethod(torch.arccos)
    expected_forward = np.array(np.arccos(_UNARY_ARC_RANGE), dtype=np.float64)
    rtol = 1e-6
    atol = 1e-6


class TestArctan(_TestUnaryFuncOpBase):
    case_name = "arctan"
    input = _UNARY_FLOAT_2x2
    forward_op = staticmethod(lucid.arctan)
    torch_forward = staticmethod(torch.atan)
    expected_forward = np.array(np.arctan(_UNARY_FLOAT_2x2), dtype=np.float64)


class TestSinh(_TestUnaryFuncOpBase):
    case_name = "sinh"
    input = _UNARY_FLOAT_2x2
    forward_op = staticmethod(lucid.sinh)
    torch_forward = staticmethod(torch.sinh)
    expected_forward = np.array(np.sinh(_UNARY_FLOAT_2x2), dtype=np.float64)
    dtype = lucid.Float64


class TestCosh(_TestUnaryFuncOpBase):
    case_name = "cosh"
    input = _UNARY_FLOAT_2x2
    forward_op = staticmethod(lucid.cosh)
    torch_forward = staticmethod(torch.cosh)
    expected_forward = np.array(np.cosh(_UNARY_FLOAT_2x2), dtype=np.float64)
    dtype = lucid.Float64


class TestTanh(_TestUnaryFuncOpBase):
    case_name = "tanh"
    input = _UNARY_FLOAT_2x2
    forward_op = staticmethod(lucid.tanh)
    torch_forward = staticmethod(torch.tanh)
    expected_forward = np.array(np.tanh(_UNARY_FLOAT_2x2), dtype=np.float64)
    rtol = 1e-6
    atol = 1e-6


class TestClip(_TestUnaryFuncOpBase):
    input: ClassVar[Any] = _UNARY_CLIP_1B
    __test__ = False

    class TestDefault(_TestUnaryFuncOpBase):
        case_name = "clip"
        input = _UNARY_CLIP_1B
        forward_op = staticmethod(lambda a: lucid.clip(a))
        torch_forward = staticmethod(
            lambda a: torch.clamp(a, min=torch.min(a), max=torch.max(a))
        )
        expected_forward = np.array(_UNARY_CLIP_1B, dtype=np.float64)

    class TestBetween(_TestUnaryFuncOpBase):
        case_name = "clip_min_max"
        input = _UNARY_CLIP_1B
        forward_op = staticmethod(
            lambda a: lucid.clip(a, min_value=-0.25, max_value=1.25)
        )
        torch_forward = staticmethod(lambda a: torch.clamp(a, min=-0.25, max=1.25))
        expected_forward = np.array([[-0.25, -0.25], [0.5, 1.25]], dtype=np.float64)

    class TestMinOnly(_TestUnaryFuncOpBase):
        case_name = "clip_min_only"
        input = _UNARY_CLIP_1B
        forward_op = staticmethod(lambda a: lucid.clip(a, min_value=-0.2))
        torch_forward = staticmethod(lambda a: torch.clamp(a, min=-0.2))
        expected_forward = np.array([[-0.2, 0.5], [1.5, 2.5]], dtype=np.float64)

    class TestMaxOnly(_TestUnaryFuncOpBase):
        case_name = "clip_max_only"
        input = _UNARY_CLIP_1B
        forward_op = staticmethod(lambda a: lucid.clip(a, max_value=1.8))
        torch_forward = staticmethod(lambda a: torch.clamp(a, max=1.8))
        expected_forward = np.array([[-1.0, 0.5], [1.5, 1.8]], dtype=np.float64)


class TestAbs(_TestUnaryFuncOpBase):
    case_name = "abs"
    input = _UNARY_NEG_2x2
    forward_op = staticmethod(lucid.abs)
    torch_forward = staticmethod(torch.abs)
    expected_forward = np.array([[1.0, 2.0], [3.0, 4.0]], dtype=np.float64)


class TestSign(_TestUnaryNoGradFuncOpBase):
    case_name = "sign_no_grad"
    input = _UNARY_NEG_2x2
    forward_op = staticmethod(lucid.sign)
    torch_forward = staticmethod(torch.sign)
    expected_forward = np.array([[-1.0, -1.0], [1.0, -1.0]], dtype=np.float64)


class TestReciprocal(_TestUnaryFuncOpBase):
    case_name = "reciprocal"
    input = [[1.0, 2.0], [4.0, 8.0]]
    forward_op = staticmethod(lucid.reciprocal)
    torch_forward = staticmethod(torch.reciprocal)
    expected_forward = np.array([[1.0, 0.5], [0.25, 0.125]], dtype=np.float64)


class TestSquare(_TestUnaryFuncOpBase):
    case_name = "square"
    input = _UNARY_FLOAT_2x2
    forward_op = staticmethod(lucid.square)
    torch_forward = staticmethod(torch.square)
    expected_forward = np.array([[1.0, 4.0], [9.0, 16.0]], dtype=np.float64)


class TestCube(_TestUnaryFuncOpBase):
    case_name = "cube"
    input = _UNARY_FLOAT_2x2
    forward_op = staticmethod(lucid.cube)
    torch_forward = staticmethod(lambda a: torch.pow(a, 3))
    expected_forward = np.array([[1.0, 8.0], [27.0, 64.0]], dtype=np.float64)


class TestTranspose(_TestUnaryFuncOpBase):
    case_name = "transpose"
    input = _UNARY_FLOAT_2x2
    forward_op = staticmethod(lambda a: lucid.transpose(a, axes=[1, 0]))
    torch_forward = staticmethod(lambda a: torch.permute(a, (1, 0)))
    expected_forward = np.array([[1.0, 3.0], [2.0, 4.0]], dtype=np.float64)


class TestT(_TestUnaryFuncOpBase):
    case_name = "T"
    input = _UNARY_FLOAT_2x2
    forward_op = staticmethod(lambda a: a.T)
    torch_forward = staticmethod(lambda a: torch.t(a))
    expected_forward = np.array([[1.0, 3.0], [2.0, 4.0]], dtype=np.float64)


class TestMT(_TestUnaryFuncOpBase):
    case_name = "mT"
    input = _UNARY_FLOAT_2x2
    forward_op = staticmethod(lambda a: a.mT)
    torch_forward = staticmethod(lambda a: a.mT)
    expected_forward = np.array([[1.0, 3.0], [2.0, 4.0]], dtype=np.float64)


class TestTrace(_TestUnaryFuncOpBase):
    case_name = "trace"
    input = _UNARY_FLOAT_2x2
    forward_op = staticmethod(lucid.trace)
    torch_forward = staticmethod(torch.trace)
    expected_forward = np.array(5.0)


class TestSum(_TestUnaryFuncOpBase):
    input: ClassVar[Any] = _UNARY_FLOAT_2x2
    __test__ = False

    class TestDefault(_TestUnaryFuncOpBase):
        case_name = "sum"
        input = _UNARY_FLOAT_2x2
        forward_op = staticmethod(lambda a: lucid.sum(a))
        torch_forward = staticmethod(lambda a: torch.sum(a))
        expected_forward = np.array(10.0)

    class TestAxisKeepDims(_TestUnaryFuncOpBase):
        case_name = "sum_axis_keepdims"
        input = _UNARY_FLOAT_2x2
        forward_op = staticmethod(lambda a: lucid.sum(a, axis=1, keepdims=True))
        torch_forward = staticmethod(lambda a: torch.sum(a, dim=1, keepdim=True))
        expected_forward = np.array([[3.0], [7.0]], dtype=np.float64)

    class TestAxisKeepDimsFalse(_TestUnaryFuncOpBase):
        case_name = "sum_axis_keepdims_false"
        input = _UNARY_FLOAT_2x2
        forward_op = staticmethod(lambda a: lucid.sum(a, axis=1, keepdims=False))
        torch_forward = staticmethod(lambda a: torch.sum(a, dim=1, keepdim=False))
        expected_forward = np.array([3.0, 7.0], dtype=np.float64)

    class TestAxisNegativeKeepDimsFalse(_TestUnaryFuncOpBase):
        case_name = "sum_axis_negative_keepdims_false"
        input = _UNARY_CUM_2x3
        forward_op = staticmethod(lambda a: lucid.sum(a, axis=-1, keepdims=False))
        torch_forward = staticmethod(lambda a: torch.sum(a, dim=-1, keepdim=False))
        expected_forward = np.array([6.0, 15.0], dtype=np.float64)


class TestMean(_TestUnaryFuncOpBase):
    input: ClassVar[Any] = _UNARY_FLOAT_2x2
    __test__ = False

    class TestDefault(_TestUnaryFuncOpBase):
        case_name = "mean"
        input = _UNARY_FLOAT_2x2
        forward_op = staticmethod(lambda a: lucid.mean(a))
        torch_forward = staticmethod(torch.mean)
        expected_forward = np.array(2.5)

    class TestAxis(_TestUnaryFuncOpBase):
        case_name = "mean_axis"
        input = _UNARY_FLOAT_2x2
        forward_op = staticmethod(lambda a: lucid.mean(a, axis=1, keepdims=False))
        torch_forward = staticmethod(lambda a: torch.mean(a, dim=1))
        expected_forward = np.array([1.5, 3.5], dtype=np.float64)

    class TestAxisKeepDims(_TestUnaryFuncOpBase):
        case_name = "mean_axis_keepdims"
        input = _UNARY_FLOAT_2x2
        forward_op = staticmethod(lambda a: lucid.mean(a, axis=1, keepdims=True))
        torch_forward = staticmethod(lambda a: torch.mean(a, dim=1, keepdim=True))
        expected_forward = np.array([[1.5], [3.5]], dtype=np.float64)

    class TestAxisNegative(_TestUnaryFuncOpBase):
        case_name = "mean_axis_negative"
        input = _UNARY_CUM_2x3
        forward_op = staticmethod(lambda a: lucid.mean(a, axis=-1, keepdims=False))
        torch_forward = staticmethod(lambda a: torch.mean(a, dim=-1, keepdim=False))
        expected_forward = np.array([2.0, 5.0], dtype=np.float64)


class TestVar(_TestUnaryFuncOpBase):
    input: ClassVar[Any] = _UNARY_FLOAT_2x2
    __test__ = False

    class TestDefault(_TestUnaryFuncOpBase):
        case_name = "var"
        input = _UNARY_FLOAT_2x2
        forward_op = staticmethod(lambda a: lucid.var(a))
        torch_forward = staticmethod(lambda a: torch.var(a, unbiased=False))
        expected_forward = np.array(1.25, dtype=np.float64)

    class TestAxis(_TestUnaryFuncOpBase):
        case_name = "var_axis"
        input = _UNARY_FLOAT_2x2
        forward_op = staticmethod(lambda a: lucid.var(a, axis=0, keepdims=False))
        torch_forward = staticmethod(lambda a: torch.var(a, dim=0, unbiased=False))
        expected_forward = np.array([1.0, 1.0], dtype=np.float64)

    class TestAxisKeepDims(_TestUnaryFuncOpBase):
        case_name = "var_axis_keepdims"
        input = _UNARY_FLOAT_2x2
        forward_op = staticmethod(lambda a: lucid.var(a, axis=1, keepdims=True))
        torch_forward = staticmethod(
            lambda a: torch.var(a, dim=1, unbiased=False, keepdim=True)
        )
        expected_forward = np.array([[0.25], [0.25]], dtype=np.float64)

    class TestAxisNegative(_TestUnaryFuncOpBase):
        case_name = "var_axis_negative"
        input = _UNARY_CUM_2x3
        forward_op = staticmethod(lambda a: lucid.var(a, axis=-1, keepdims=False))
        torch_forward = staticmethod(
            lambda a: torch.var(a, dim=-1, unbiased=False, keepdim=False)
        )
        expected_forward = np.array([2 / 3, 2 / 3], dtype=np.float64)


class TestMin(_TestUnaryFuncOpBase):
    input: ClassVar[Any] = _UNARY_FLOAT_2x2
    __test__ = False

    class TestDefault(_TestUnaryFuncOpBase):
        case_name = "min"
        input = _UNARY_FLOAT_2x2
        forward_op = staticmethod(lucid.min)
        torch_forward = staticmethod(torch.amin)
        expected_forward = np.array(1.0)

    class TestAxis(_TestUnaryFuncOpBase):
        case_name = "min_axis"
        input = _UNARY_FLOAT_2x2
        forward_op = staticmethod(lambda a: lucid.min(a, axis=1, keepdims=True))
        torch_forward = staticmethod(lambda a: torch.amin(a, dim=1, keepdim=True))
        expected_forward = np.array([[1.0], [3.0]], dtype=np.float64)

    class TestAxisKeepDimsFalse(_TestUnaryFuncOpBase):
        case_name = "min_axis_keepdims_false"
        input = _UNARY_FLOAT_2x2
        forward_op = staticmethod(lambda a: lucid.min(a, axis=0, keepdims=False))
        torch_forward = staticmethod(lambda a: torch.amin(a, dim=0, keepdim=False))
        expected_forward = np.array([1.0, 2.0], dtype=np.float64)

    class TestAxisNegative(_TestUnaryFuncOpBase):
        case_name = "min_axis_negative"
        input = _UNARY_CUM_2x3
        forward_op = staticmethod(lambda a: lucid.min(a, axis=-1, keepdims=False))
        torch_forward = staticmethod(lambda a: torch.amin(a, dim=-1, keepdim=False))
        expected_forward = np.array([1.0, 4.0], dtype=np.float64)


class TestMax(_TestUnaryFuncOpBase):
    input: ClassVar[Any] = _UNARY_FLOAT_2x2
    __test__ = False

    class TestDefault(_TestUnaryFuncOpBase):
        case_name = "max"
        input = _UNARY_FLOAT_2x2
        forward_op = staticmethod(lucid.max)
        torch_forward = staticmethod(torch.amax)
        expected_forward = np.array(4.0)

    class TestAxis(_TestUnaryFuncOpBase):
        case_name = "max_axis"
        input = _UNARY_FLOAT_2x2
        forward_op = staticmethod(lambda a: lucid.max(a, axis=1, keepdims=False))
        torch_forward = staticmethod(lambda a: torch.amax(a, dim=1))
        expected_forward = np.array([2.0, 4.0], dtype=np.float64)

    class TestAxisKeepDims(_TestUnaryFuncOpBase):
        case_name = "max_axis_keepdims"
        input = _UNARY_FLOAT_2x2
        forward_op = staticmethod(lambda a: lucid.max(a, axis=0, keepdims=True))
        torch_forward = staticmethod(lambda a: torch.amax(a, dim=0, keepdim=True))
        expected_forward = np.array([[3.0, 4.0]], dtype=np.float64)

    class TestAxisNegative(_TestUnaryFuncOpBase):
        case_name = "max_axis_negative"
        input = _UNARY_CUM_2x3
        forward_op = staticmethod(lambda a: lucid.max(a, axis=-1, keepdims=False))
        torch_forward = staticmethod(lambda a: torch.amax(a, dim=-1, keepdim=False))
        expected_forward = np.array([3.0, 6.0], dtype=np.float64)


class TestSwapaxes(_TestUnaryFuncOpBase):
    case_name = "swapaxes"
    input = _UNARY_FLOAT_2x2
    forward_op = staticmethod(lambda a: lucid.swapaxes(a, axis1=0, axis2=1))
    torch_forward = staticmethod(lambda a: torch.swapaxes(a, 0, 1))
    expected_forward = np.array([[1.0, 3.0], [2.0, 4.0]], dtype=np.float64)


class TestTransposeDefaultAxes(_TestUnaryFuncOpBase):
    case_name = "transpose_default_axes"
    input = _UNARY_RECT_2x3
    forward_op = staticmethod(lambda a: lucid.transpose(a, axes=None))
    torch_forward = staticmethod(lambda a: torch.permute(a, (1, 0)))
    expected_forward = np.array([[1.0, 4.0], [2.0, 5.0], [3.0, 6.0]], dtype=np.float64)


class TestRound(_TestUnaryNoGradFuncOpBase):
    input: ClassVar[Any] = _UNARY_DECIMALS
    __test__ = False

    class TestDefault(_TestUnaryNoGradFuncOpBase):
        case_name = "round_no_grad"
        input = [[1.2, 2.7], [3.9, 4.0]]
        forward_op = staticmethod(lambda a: lucid.round(a, decimals=0))
        torch_forward = staticmethod(lambda a: torch.round(a))
        expected_forward = np.array([[1.0, 3.0], [4.0, 4.0]], dtype=np.float64)

    class TestDecimals(_TestUnaryNoGradFuncOpBase):
        case_name = "round_decimals_no_grad"
        input = _UNARY_DECIMALS
        forward_op = staticmethod(lambda a: lucid.round(a, decimals=1))
        torch_forward = staticmethod(lambda a: torch.round(a, decimals=1))
        expected_forward = np.array([[1.1, 2.7], [3.4, 5.0]], dtype=np.float64)


class TestFloor(_TestUnaryNoGradFuncOpBase):
    case_name = "floor_no_grad"
    input = [[1.2, 2.7], [3.9, 4.1]]
    forward_op = staticmethod(lucid.floor)
    torch_forward = staticmethod(torch.floor)
    expected_forward = np.array([[1.0, 2.0], [3.0, 4.0]], dtype=np.float64)


class TestCeil(_TestUnaryNoGradFuncOpBase):
    case_name = "ceil_no_grad"
    input = [[1.2, 2.7], [3.1, 4.0]]
    forward_op = staticmethod(lucid.ceil)
    torch_forward = staticmethod(torch.ceil)
    expected_forward = np.array([[2.0, 3.0], [4.0, 4.0]], dtype=np.float64)


class TestCumprod(_TestUnaryFuncOpBase):
    input: ClassVar[Any] = _UNARY_CUM_2x3
    __test__ = False

    class TestDefault(_TestUnaryFuncOpBase):
        case_name = "cumprod"
        input = _UNARY_CUM_2x3
        forward_op = staticmethod(lambda a: lucid.cumprod(a, axis=1))
        torch_forward = staticmethod(lambda a: torch.cumprod(a, dim=1))
        expected_forward = np.array(
            [[1.0, 2.0, 6.0], [4.0, 20.0, 120.0]], dtype=np.float64
        )

    class TestDefaultAxis(_TestUnaryFuncOpBase):
        case_name = "cumprod_default_axis"
        input = _UNARY_CUM_2x3
        forward_op = staticmethod(lucid.cumprod)
        torch_forward = staticmethod(lambda a: torch.cumprod(a, dim=-1))
        expected_forward = np.array(
            [[1.0, 2.0, 6.0], [4.0, 20.0, 120.0]], dtype=np.float64
        )

    class TestAxisZero(_TestUnaryFuncOpBase):
        case_name = "cumprod_axis_zero"
        input = _UNARY_CUM_2x3
        forward_op = staticmethod(lambda a: lucid.cumprod(a, axis=0))
        torch_forward = staticmethod(lambda a: torch.cumprod(a, dim=0))
        expected_forward = np.array(
            [[1.0, 2.0, 3.0], [4.0, 10.0, 18.0]], dtype=np.float64
        )

    class TestNegativeAxis(_TestUnaryFuncOpBase):
        case_name = "cumprod_negative_axis"
        input = _UNARY_CUM_2x3
        forward_op = staticmethod(lambda a: lucid.cumprod(a, axis=-1))
        torch_forward = staticmethod(lambda a: torch.cumprod(a, dim=-1))
        expected_forward = np.array(
            [[1.0, 2.0, 6.0], [4.0, 20.0, 120.0]], dtype=np.float64
        )


class TestCumsum(_TestUnaryFuncOpBase):
    input: ClassVar[Any] = _UNARY_CUM_2x3
    __test__ = False

    class TestDefault(_TestUnaryFuncOpBase):
        case_name = "cumsum"
        input = _UNARY_CUM_2x3
        forward_op = staticmethod(lambda a: lucid.cumsum(a, axis=1))
        torch_forward = staticmethod(lambda a: torch.cumsum(a, dim=1))
        expected_forward = np.array(
            [[1.0, 3.0, 6.0], [4.0, 9.0, 15.0]], dtype=np.float64
        )

    class TestDefaultAxis(_TestUnaryFuncOpBase):
        case_name = "cumsum_default_axis"
        input = _UNARY_CUM_2x3
        forward_op = staticmethod(lucid.cumsum)
        torch_forward = staticmethod(lambda a: torch.cumsum(a, dim=-1))
        expected_forward = np.array(
            [[1.0, 3.0, 6.0], [4.0, 9.0, 15.0]], dtype=np.float64
        )

    class TestAxisZero(_TestUnaryFuncOpBase):
        case_name = "cumsum_axis_zero"
        input = _UNARY_CUM_2x3
        forward_op = staticmethod(lambda a: lucid.cumsum(a, axis=0))
        torch_forward = staticmethod(lambda a: torch.cumsum(a, dim=0))
        expected_forward = np.array(
            [[1.0, 2.0, 3.0], [5.0, 7.0, 9.0]], dtype=np.float64
        )

    class TestNegativeAxis(_TestUnaryFuncOpBase):
        case_name = "cumsum_negative_axis"
        input = _UNARY_CUM_2x3
        forward_op = staticmethod(lambda a: lucid.cumsum(a, axis=-1))
        torch_forward = staticmethod(lambda a: torch.cumsum(a, dim=-1))
        expected_forward = np.array(
            [[1.0, 3.0, 6.0], [4.0, 9.0, 15.0]], dtype=np.float64
        )



import math
from dataclasses import dataclass
from functools import partial
from typing import Callable, ClassVar, Literal, NamedTuple

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid import register_model
from lucid._tensor import Tensor

__all__ = [
    "CSPNet",
    "CSPNetConfig",
    "csp_resnet_50",
    "csp_resnext_50_32x4d",
    "csp_darknet_53",
]


def _normalize_stage_specs(
    values: (
        tuple[tuple[object, ...], ...] | list[tuple[object, ...]] | list[list[object]]
    ),
) -> tuple[tuple[int, int, bool], ...]:
    normalized = tuple(tuple(spec) for spec in values)
    if len(normalized) == 0:
        raise ValueError("stage_specs must contain at least one stage spec")
    for spec in normalized:
        if len(spec) != 3:
            raise ValueError(
                "each stage spec must contain exactly 3 values: (stage_width, num_layers, downsample)"
            )
        stage_width, num_layers, downsample = spec
        if not isinstance(stage_width, int) or stage_width <= 0:
            raise ValueError("stage_width values must be positive integers")
        if not isinstance(num_layers, int) or num_layers <= 0:
            raise ValueError("num_layers values must be positive integers")
        if not isinstance(downsample, bool):
            raise TypeError("downsample values must be booleans")
    return normalized  # type: ignore[return-value]


@dataclass
class CSPNetConfig:
    stage_specs: (
        tuple[tuple[object, ...], ...] | list[tuple[object, ...]] | list[list[object]]
    )
    stack_type: Literal["resnet", "resnext", "darknet"]
    in_channels: int = 3
    stem_channels: int = 64
    num_classes: int = 1000
    norm: Callable[..., nn.Module] = nn.BatchNorm2d
    act: Callable[..., nn.Module] = nn.ReLU
    split_ratio: float = 0.5
    global_pool: Literal["avg", "max"] = "avg"
    dropout: float = 0.0
    feature_channels: int | None = None
    pre_kernel_size: int = 1
    groups: int = 1
    base_width: int = 64

    def __post_init__(self) -> None:
        self.stage_specs = _normalize_stage_specs(self.stage_specs)
        if self.stack_type not in {"resnet", "resnext", "darknet"}:
            raise ValueError(
                "stack_type must be one of 'resnet', 'resnext', or 'darknet'"
            )
        if self.in_channels <= 0:
            raise ValueError("in_channels must be greater than 0")
        if self.stem_channels <= 0:
            raise ValueError("stem_channels must be greater than 0")
        if self.num_classes <= 0:
            raise ValueError("num_classes must be greater than 0")
        if not callable(self.norm):
            raise TypeError("norm must be callable")
        if not callable(self.act):
            raise TypeError("act must be callable")
        if self.split_ratio <= 0 or self.split_ratio >= 1:
            raise ValueError("split_ratio must be in the range (0, 1)")
        if self.global_pool not in {"avg", "max"}:
            raise ValueError("global_pool must be either 'avg' or 'max'")
        if self.dropout < 0 or self.dropout >= 1:
            raise ValueError("dropout must be in the range [0, 1)")
        if self.feature_channels is not None and self.feature_channels <= 0:
            raise ValueError("feature_channels must be greater than 0 when provided")
        if self.pre_kernel_size <= 0:
            raise ValueError("pre_kernel_size must be greater than 0")
        if self.groups <= 0:
            raise ValueError("groups must be greater than 0")
        if self.base_width <= 0:
            raise ValueError("base_width must be greater than 0")


class _ConvBNAct(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        k: int = 3,
        s: int = 1,
        p: int | None = None,
        groups: int = 1,
        bias: bool = False,
        norm: type[nn.Module] = nn.BatchNorm2d,
        act: type[nn.Module] = nn.ReLU,
    ) -> None:
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=k,
            stride=s,
            padding=p if p is not None else "same",
            groups=groups,
            bias=bias,
        )
        self.norm = norm(out_channels)
        self.act = act()

    def forward(self, x: Tensor) -> Tensor:
        return self.act(self.norm(self.conv(x)))


class _Bottleneck(nn.Module):
    expansion: ClassVar[int] = 4

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int = 1,
        downsample: nn.Module | None = None,
        norm: type[nn.Module] = nn.BatchNorm2d,
        act: type[nn.Module] = nn.ReLU,
        groups: int = 1,
        base_width: int = 64,
        **kwargs,
    ) -> None:
        super().__init__()
        width = int(out_channels * (base_width / 64.0)) * groups

        self.conv1 = _ConvBNAct(in_channels, width, k=1, s=1, p=0, norm=norm, act=act)
        self.conv2 = _ConvBNAct(
            width, width, k=3, s=stride, p=1, groups=groups, norm=norm, act=act
        )
        self.conv3 = nn.Conv2d(
            width, out_channels * self.expansion, kernel_size=1, bias=False
        )
        self.norm3 = norm(out_channels * self.expansion)
        self.act = act()
        self.downsample = downsample

    def forward(self, x: Tensor) -> Tensor:
        identity = x
        out = self.conv1(x)
        out = self.conv2(out)
        out = self.norm3(self.conv3(out))

        if self.downsample is not None:
            identity = self.downsample(x)
        out += identity
        return self.act(out)


class _DarknetBottleneck(nn.Module):
    expansion: ClassVar[int] = 1

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int = 1,
        norm: type[nn.Module] = nn.BatchNorm2d,
        act: type[nn.Module] = partial(nn.LeakyReLU, negative_slope=0.1),
        **kwargs,
    ) -> None:
        super().__init__()
        self.conv1 = _ConvBNAct(
            in_channels, out_channels, k=1, s=1, p=0, norm=norm, act=act
        )
        self.conv2 = _ConvBNAct(
            out_channels, out_channels, k=3, s=stride, p=1, norm=norm, act=act
        )

        self.use_proj = not (stride == 1 and in_channels == out_channels)
        self.downsample = (
            nn.Identity()
            if not self.use_proj
            else nn.Sequential(
                nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False),
                norm(out_channels),
            )
        )
        self.act = act()

    def forward(self, x: Tensor) -> Tensor:
        identity = x
        out = self.conv1(x)
        out = self.conv2(out)
        if self.use_proj:
            identity = self.downsample(x)

        out += identity
        return self.act(out)


class _StackOut(NamedTuple):
    module: nn.Module
    out_channels: int


def _resnet_stack_factory(
    block_cls: type[nn.Module],
    norm: type[nn.Module] = nn.BatchNorm2d,
    act: type[nn.Module] = nn.ReLU,
    groups: int = 1,
    base_width: int = 64,
    **kwargs,
) -> Callable[[int, int, int], _StackOut]:
    expansion = getattr(block_cls, "expansion", 1)

    def make_stack(
        in_channels: int, num_layers: int, stride_first: int = 1
    ) -> _StackOut:
        channels = math.ceil(in_channels / expansion)
        layers = []

        down = None
        if in_channels != channels * expansion:
            down = nn.Sequential(
                nn.Conv2d(in_channels, channels * expansion, kernel_size=1, bias=False),
                norm(channels * expansion),
            )

        base_kwargs = dict(norm=norm, act=act, groups=groups, base_width=base_width)
        layers.append(
            block_cls(
                in_channels,
                channels,
                stride=stride_first,
                downsample=down,
                **base_kwargs,
            )
        )
        in_channels = channels * expansion
        for _ in range(1, num_layers):
            layers.append(
                block_cls(
                    in_channels, channels, stride=1, downsample=None, **base_kwargs
                )
            )

        return _StackOut(nn.Sequential(*layers), in_channels)

    make_stack.required_multiple = expansion
    return make_stack


def _darknet_stack_factory(
    block_cls: type[nn.Module],
    norm: type[nn.Module] = nn.BatchNorm2d,
    act: type[nn.Module] = partial(nn.LeakyReLU, negative_slope=0.1),
    **kwargs,
) -> Callable[[int, int, int], _StackOut]:
    expansion = getattr(block_cls, "expansion", 1)

    def make_stack(
        in_channels: int, num_layers: int, stride_first: int = 1
    ) -> _StackOut:
        channels = in_channels // expansion
        base_kwargs = dict(downsample=None, norm=norm, act=act)

        layers = []
        layers.append(
            block_cls(
                in_channels, channels * expansion, stride=stride_first, **base_kwargs
            )
        )
        in_ch = channels * expansion
        for _ in range(1, num_layers):
            layers.append(
                block_cls(in_ch, channels * expansion, stride=1, **base_kwargs)
            )

        return _StackOut(nn.Sequential(*layers), in_ch)

    make_stack.required_multiple = expansion
    return make_stack


class _CSPStage(nn.Module):
    def __init__(
        self,
        in_channels: int,
        stage_width: int,
        num_layers: int,
        block_stack_fn: Callable[[int, int, int], _StackOut],
        split_ratio: float = 0.5,
        downsample: bool = False,
        norm: type[nn.Module] = nn.BatchNorm2d,
        act: type[nn.Module] = nn.ReLU,
        pre_kernel_size: int = 1,
    ) -> None:
        super().__init__()
        self.pre = _ConvBNAct(
            in_channels,
            stage_width,
            k=pre_kernel_size,
            s=(2 if downsample else 1),
            p=(pre_kernel_size - 1) // 2,
            norm=norm,
            act=act,
        )

        c1 = int(round(stage_width * split_ratio))
        c2 = stage_width - c1
        assert c1 > 0 and c2 > 0

        req = getattr(block_stack_fn, "required_multiple", 1)
        if c2 % req != 0:
            c2 = max(req, (c2 // req) * req)
            c1 = stage_width - c2
        self.c1, self.c2 = c1, c2

        self.part1_proj = _ConvBNAct(stage_width, c1, k=1, s=1, p=0, norm=norm, act=act)
        self.part2_proj = _ConvBNAct(stage_width, c2, k=1, s=1, p=0, norm=norm, act=act)

        stack_out = block_stack_fn(c2, num_layers, stride_first=1)
        self.block_stack = stack_out.module
        self.block_out = stack_out.out_channels

        self.merge = _ConvBNAct(
            c1 + self.block_out, stage_width, k=1, s=1, p=0, norm=norm, act=act
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.pre(x)
        y1 = self.part1_proj(x)
        y2 = self.part2_proj(x)
        y2 = self.block_stack(y2)

        y = lucid.concatenate([y1, y2], axis=1)
        y = self.merge(y)
        return y


class CSPNet(nn.Module):
    def __init__(self, config: CSPNetConfig) -> None:
        super().__init__()
        self.config = config

        if config.stack_type == "darknet":
            block_stack_fn = _darknet_stack_factory(
                _DarknetBottleneck,
                norm=config.norm,
                act=config.act,
            )
        else:
            block_stack_fn = _resnet_stack_factory(
                _Bottleneck,
                norm=config.norm,
                act=config.act,
                groups=config.groups,
                base_width=config.base_width,
            )

        self.stem = nn.Sequential(
            _ConvBNAct(
                config.in_channels,
                config.stem_channels,
                k=3,
                s=2,
                norm=config.norm,
                act=config.act,
            ),
            _ConvBNAct(
                config.stem_channels,
                config.stem_channels,
                k=3,
                s=1,
                norm=config.norm,
                act=config.act,
            ),
        )

        stages = []
        in_ch = config.stem_channels
        for stage_width, num_layers, downsample in config.stage_specs:
            stages.append(
                _CSPStage(
                    in_ch,
                    stage_width,
                    num_layers,
                    block_stack_fn,
                    split_ratio=config.split_ratio,
                    downsample=downsample,
                    norm=config.norm,
                    act=config.act,
                    pre_kernel_size=config.pre_kernel_size,
                )
            )
            in_ch = stage_width
        self.stages = nn.Sequential(*stages)

        if config.feature_channels is not None and config.feature_channels != in_ch:
            self.pre_head = _ConvBNAct(
                in_ch,
                config.feature_channels,
                k=1,
                s=1,
                p=0,
                norm=config.norm,
                act=config.act,
            )
            in_ch = config.feature_channels
        else:
            self.pre_head = nn.Identity()

        self.num_classes = config.num_classes
        self.head_pool = (
            nn.AdaptiveAvgPool2d((1, 1))
            if config.global_pool == "avg"
            else nn.AdaptiveMaxPool2d((1, 1))
        )
        self.head = nn.Sequential(
            nn.Flatten(),
            nn.Dropout(config.dropout) if config.dropout > 0.0 else nn.Identity(),
            nn.Linear(in_ch, config.num_classes),
        )

        self.apply(self.init_weights)

    def init_weights(self, m: nn.Module):
        if isinstance(m, nn.Conv2d):
            nn.init.kaiming_normal(m.weight, mode="fan_out")
        elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
            nn.init.constant(m.weight, 1.0)
            nn.init.constant(m.bias, 0.0)

    def forward_features(
        self, x: Tensor, return_stage_out: bool = False
    ) -> Tensor | list[Tensor]:
        feats = []
        x = self.stem(x)
        for stage in self.stages:
            x = stage(x)
            if return_stage_out:
                feats.append(x)

        return feats if return_stage_out else x

    def forward(self, x: Tensor) -> Tensor:
        x = self.forward_features(x, return_stage_out=False)
        x = self.pre_head(x)
        x = self.head_pool(x)
        x = self.head(x)
        return x


def _raise_for_locked_factory_kwargs(
    kwargs: dict[str, object],
    locked_fields: set[str],
    message: str,
) -> None:
    if locked_fields & kwargs.keys():
        raise TypeError(message)


def _build_cspnet_config(
    *,
    stage_specs: tuple[tuple[int, int, bool], ...],
    stack_type: Literal["resnet", "resnext", "darknet"],
    num_classes: int,
    stem_channels: int,
    split_ratio: float,
    **kwargs,
) -> CSPNetConfig:
    return CSPNetConfig(
        stage_specs=stage_specs,
        stack_type=stack_type,
        num_classes=num_classes,
        stem_channels=stem_channels,
        split_ratio=split_ratio,
        **kwargs,
    )


@register_model
def csp_resnet_50(
    num_classes: int = 1000, split_ratio: float = 0.5, stem_channels: int = 64, **kwargs
) -> CSPNet:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"stage_specs", "stack_type", "groups", "base_width", "feature_channels"},
        "factory variants do not allow overriding preset stage_specs, stack_type, groups, base_width, or feature_channels",
    )
    config = _build_cspnet_config(
        stage_specs=(
            (256, 3, False),
            (512, 4, True),
            (1024, 6, True),
            (2048, 3, True),
        ),
        stack_type="resnet",
        num_classes=num_classes,
        stem_channels=stem_channels,
        split_ratio=split_ratio,
        groups=1,
        base_width=64,
        feature_channels=1024,
        **kwargs,
    )
    return CSPNet(config)


@register_model
def csp_resnext_50_32x4d(
    num_classes: int = 1000, split_ratio: float = 0.5, stem_channels: int = 64, **kwargs
) -> CSPNet:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"stage_specs", "stack_type", "groups", "base_width", "feature_channels"},
        "factory variants do not allow overriding preset stage_specs, stack_type, groups, base_width, or feature_channels",
    )
    config = _build_cspnet_config(
        stage_specs=(
            (256, 3, False),
            (512, 4, True),
            (1024, 6, True),
            (2048, 3, True),
        ),
        stack_type="resnext",
        num_classes=num_classes,
        stem_channels=stem_channels,
        split_ratio=split_ratio,
        groups=32,
        base_width=4,
        feature_channels=1024,
        **kwargs,
    )
    return CSPNet(config)


@register_model
def csp_darknet_53(
    num_classes: int = 1000, split_ratio: float = 0.5, stem_channels: int = 32, **kwargs
) -> CSPNet:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"stage_specs", "stack_type", "feature_channels", "pre_kernel_size"},
        "factory variants do not allow overriding preset stage_specs, stack_type, feature_channels, or pre_kernel_size",
    )
    config = _build_cspnet_config(
        stage_specs=(
            (64, 1, True),
            (128, 2, True),
            (256, 8, True),
            (512, 8, True),
            (1024, 4, True),
        ),
        stack_type="darknet",
        num_classes=num_classes,
        stem_channels=stem_channels,
        split_ratio=split_ratio,
        feature_channels=1024,
        pre_kernel_size=3,
        **kwargs,
    )
    return CSPNet(config)



from dataclasses import dataclass
from typing import Type
import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor
from lucid.models.base import PreTrainedModelMixin

__all__ = ["LeNet", "LeNetConfig", "lenet_1", "lenet_4", "lenet_5"]


@dataclass
class LeNetConfig:
    conv_layers: list[dict[str, int]]
    clf_layers: list[int]
    clf_in_features: int
    base_activation: Type[nn.Module] = nn.Tanh

    def __post_init__(self) -> None:
        if len(self.conv_layers) != 2:
            raise ValueError("LeNet requires exactly 2 convolution layer configs.")

        for idx, layer_config in enumerate(self.conv_layers, start=1):
            if "out_channels" not in layer_config:
                raise ValueError(
                    f"LeNet conv layer config at index {idx} must define "
                    "'out_channels'."
                )

        if not self.clf_layers:
            raise ValueError("LeNet requires at least 1 classifier layer.")

        if self.clf_in_features <= 0:
            raise ValueError("LeNet clf_in_features must be greater than 0.")


class LeNet(nn.Module, PreTrainedModelMixin):
    def __init__(self, config: LeNetConfig) -> None:
        super().__init__()
        self.config = config

        self.conv1 = nn.Sequential(
            nn.Conv2d(1, config.conv_layers[0]["out_channels"], kernel_size=5),
            config.base_activation(),
            nn.AvgPool2d(2, 2),
        )
        self.conv2 = nn.Sequential(
            nn.Conv2d(
                config.conv_layers[0]["out_channels"],
                config.conv_layers[1]["out_channels"],
                kernel_size=5,
            ),
            config.base_activation(),
            nn.AvgPool2d(2, 2),
        )

        in_features = config.clf_in_features
        n_clf_layers = len(config.clf_layers)
        for idx, units in enumerate(config.clf_layers, start=1):
            self.add_module(f"fc{idx}", nn.Linear(in_features, units))
            if idx < n_clf_layers:
                self.add_module(f"tanh{idx + 2}", config.base_activation())
            in_features = units

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv1(x)
        x = self.conv2(x)
        x = x.reshape(x.shape[0], -1)

        idx = 1
        while hasattr(self, f"fc{idx}"):
            x = getattr(self, f"fc{idx}")(x)
            if hasattr(self, f"tanh{idx + 2}"):
                x = getattr(self, f"tanh{idx + 2}")(x)
            idx += 1

        return x


def _lenet_config_from_kwargs(
    *,
    conv_layers: list[dict[str, int]],
    clf_layers: list[int],
    clf_in_features: int,
    **kwargs,
) -> LeNetConfig:
    legacy_base_activation = kwargs.pop("_base_activation", None)
    if legacy_base_activation is not None:
        if "base_activation" in kwargs:
            raise TypeError(
                "lenet factory received both '_base_activation' and "
                "'base_activation'."
            )
        kwargs["base_activation"] = legacy_base_activation

    return LeNetConfig(
        conv_layers=conv_layers,
        clf_layers=clf_layers,
        clf_in_features=clf_in_features,
        **kwargs,
    )


@register_model
def lenet_1(**kwargs) -> LeNet:
    config = _lenet_config_from_kwargs(
        conv_layers=[{"out_channels": 4}, {"out_channels": 12}],
        clf_layers=[10],
        clf_in_features=12 * 4 * 4,
        **kwargs,
    )
    return LeNet(config)


@register_model
def lenet_4(**kwargs) -> LeNet:
    config = _lenet_config_from_kwargs(
        conv_layers=[{"out_channels": 4}, {"out_channels": 12}],
        clf_layers=[84, 10],
        clf_in_features=12 * 4 * 4,
        **kwargs,
    )
    return LeNet(config)


@register_model
def lenet_5(**kwargs) -> LeNet:
    config = _lenet_config_from_kwargs(
        conv_layers=[{"out_channels": 6}, {"out_channels": 16}],
        clf_layers=[120, 84, 10],
        clf_in_features=16 * 5 * 5,
        **kwargs,
    )
    return LeNet(config)



from typing import Sequence
import random
import math

import lucid

from ._base import Dataset, Subset

__all__ = ["random_split"]


def _resolve_lengths_from_fractions(fractions: Sequence[float], n: int) -> list[int]:
    if not fractions:
        raise ValueError("fractions must be non-empty.")
    if any(f < 0 for f in fractions):
        raise ValueError("Fractional lengths mus be non-negative.")

    s = sum(fractions)
    if not math.isclose(s, 1.0, rel_tol=1e-6, abs_tol=1e-6):
        raise ValueError(f"When passing fractions, they must sum to 1.0 (got {s}).")

    base = [int(math.floor(f * n)) for f in fractions]
    remainder = n - sum(base)
    for i in range(remainder):
        base[i % len(base)] += 1

    return base


def random_split(
    dataset: Dataset, lengths: Sequence[int | float], seed: int | None = None
) -> tuple[Subset, ...]:
    n = len(dataset)
    if not lengths:
        raise ValueError("lengths must be non-empty.")

    all_int = all(isinstance(l, int) for l in lengths)
    all_float = all(isinstance(l, float) for l in lengths)

    if not (all_int or all_float):
        return TypeError("lengths must be all integers or all floats.")

    if all_float:
        int_lengths = _resolve_lengths_from_fractions(lengths, n)
    else:
        int_lengths = list(lengths)
        s = sum(int_lengths)
        if s != n:
            raise ValueError(
                f"Sum of input lengths ({s}) does not equal dataset length ({n})."
            )
        if any(l < 0 for l in int_lengths):
            raise ValueError("All split lengths must be non-negative.")

    if seed is None:
        seed = lucid.random.get_seed()
    rng = random.Random(seed)

    indices = list(range(n))
    rng.shuffle(indices)

    splits: list[Subset] = []
    offset = 0
    for length in int_lengths:
        split_idx = indices[offset : offset + length]
        splits.append(Subset(dataset, split_idx))
        offset += length

    return tuple(splits)



import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor

__all__ = ["Embedding"]


class Embedding(nn.Module):
    def __init__(
        self,
        num_embeddings: int,
        embedding_dim: int,
        padding_idx: int | None = None,
        max_norm: float | None = None,
        norm_type: float = 2.0,
        _weight: Tensor | None = None,
    ) -> None:
        super().__init__()
        self.num_embeddings = num_embeddings
        self.embedding_dim = embedding_dim

        if padding_idx is None:
            self.padding_idx = None
        else:
            pad = int(padding_idx)
            if pad < 0:
                pad += num_embeddings
            if pad < 0 or pad >= num_embeddings:
                raise IndexError("padding_idx out of range.")
            self.padding_idx = pad

        self.max_norm = max_norm
        self.norm_type = norm_type

        if _weight is None:
            self.weight = nn.Parameter(
                lucid.random.uniform(-0.1, 0.1, (num_embeddings, embedding_dim))
            )
        else:
            self.weight = nn.Parameter(_weight)

        if self.padding_idx is not None:
            self.weight.data[self.padding_idx] = 0

    def forward(self, input_: Tensor) -> Tensor:
        return F.embedding(
            input_, self.weight, self.padding_idx, self.max_norm, self.norm_type
        )

    def reset_parameters(self) -> None:
        self.weight.data = lucid.random.uniform(-0.1, 0.1, self.weight.shape)
        if self.padding_idx is not None:
            self.weight.data[self.padding_idx] = 0



from dataclasses import dataclass
from functools import partial
from typing import Type

import lucid
import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor

__all__ = [
    "InceptionNeXt",
    "InceptionNeXtConfig",
    "inception_next_atto",
    "inception_next_tiny",
    "inception_next_small",
    "inception_next_base",
]


def _normalize_positive_int_sequence(
    values: tuple[int, ...] | list[int],
    name: str,
) -> tuple[int, ...]:
    normalized = tuple(values)
    if len(normalized) == 0:
        raise ValueError(f"{name} must contain at least one stage value")
    if any(not isinstance(value, int) or value <= 0 for value in normalized):
        raise ValueError(f"{name} values must be positive integers")
    return normalized


@dataclass
class InceptionNeXtConfig:
    num_classes: int = 1000
    depths: tuple[int, ...] | list[int] = (3, 3, 9, 3)
    dims: tuple[int, ...] | list[int] = (96, 192, 384, 768)
    token_mixers: object | None = None
    mlp_ratios: int | tuple[int, ...] | list[int] = (4, 4, 4, 3)
    head_fn: object | None = None
    drop_rate: float = 0.0
    drop_path_rate: float = 0.0
    ls_init_value: float = 1e-6

    def __post_init__(self) -> None:
        self.depths = _normalize_positive_int_sequence(self.depths, "depths")
        self.dims = _normalize_positive_int_sequence(self.dims, "dims")
        num_stage = len(self.depths)

        if len(self.dims) != num_stage:
            raise ValueError("dims must have the same number of stages as depths")
        if self.num_classes <= 0:
            raise ValueError("num_classes must be greater than 0")

        if self.token_mixers is None:
            self.token_mixers = _InceptionDWConv2d

        if isinstance(self.token_mixers, (list, tuple)):
            if len(self.token_mixers) != num_stage:
                raise ValueError(
                    "token_mixers must have the same number of stages as depths"
                )
            if any(not callable(token_mixer) for token_mixer in self.token_mixers):
                raise TypeError("token_mixers entries must be callable")
            self.token_mixers = tuple(self.token_mixers)
        elif callable(self.token_mixers):
            self.token_mixers = tuple(self.token_mixers for _ in range(num_stage))
        else:
            raise TypeError("token_mixers must be callable or a sequence of callables")

        if isinstance(self.mlp_ratios, (list, tuple)):
            if len(self.mlp_ratios) != num_stage:
                raise ValueError(
                    "mlp_ratios must have the same number of stages as depths"
                )
            if any(
                not isinstance(mlp_ratio, int) or mlp_ratio <= 0
                for mlp_ratio in self.mlp_ratios
            ):
                raise ValueError("mlp_ratios values must be positive integers")
            self.mlp_ratios = tuple(self.mlp_ratios)
        elif isinstance(self.mlp_ratios, int) and self.mlp_ratios > 0:
            self.mlp_ratios = tuple(self.mlp_ratios for _ in range(num_stage))
        else:
            raise ValueError(
                "mlp_ratios must be a positive integer or sequence of positive integers"
            )

        if self.head_fn is None:
            self.head_fn = _MLPHead
        if not callable(self.head_fn):
            raise TypeError("head_fn must be callable")
        if self.drop_rate < 0 or self.drop_rate >= 1:
            raise ValueError("drop_rate must be in the range [0, 1)")
        if self.drop_path_rate < 0 or self.drop_path_rate > 1:
            raise ValueError("drop_path_rate must be in the range [0, 1]")
        if self.ls_init_value < 0:
            raise ValueError("ls_init_value must be greater than or equal to 0")


class _InceptionDWConv2d(nn.Module):
    def __init__(
        self,
        in_channels: int,
        square_kernel_size: int = 3,
        band_kernel_size: int = 11,
        branch_ratio: float = 0.25,
    ) -> None:
        super().__init__()
        gc = int(in_channels * branch_ratio)

        self.dwconv_hw = nn.Conv2d(
            gc, gc, kernel_size=square_kernel_size, padding="same", groups=gc
        )
        self.dwconv_w = nn.Conv2d(
            gc, gc, kernel_size=(1, band_kernel_size), padding="same", groups=gc
        )
        self.dwconv_h = nn.Conv2d(
            gc, gc, kernel_size=(band_kernel_size, 1), padding="same", groups=gc
        )
        self.split_indices = (in_channels - 3 * gc, gc, gc, gc)

    def forward(self, x: Tensor) -> Tensor:
        x_id, x_hw, x_w, x_h = x.split(self.split_indices, axis=1)
        return lucid.concatenate(
            (x_id, self.dwconv_hw(x_hw), self.dwconv_w(x_w), self.dwconv_h(x_h)),
            axis=1,
        )


class _ConvMLP(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_features: int | None = None,
        out_features: int | None = None,
        bias: bool = True,
        act_layer: Type[nn.Module] = nn.ReLU,
        norm_layer: Type[nn.Module] | None = None,
        drop: float = 0.0,
    ) -> None:
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features

        self.fc1 = nn.Conv2d(in_features, hidden_features, kernel_size=1, bias=bias)
        self.norm = norm_layer(hidden_features) if norm_layer else nn.Identity()
        self.act = act_layer()

        self.drop = nn.Dropout(drop)
        self.fc2 = nn.Conv2d(hidden_features, out_features, kernel_size=1, bias=bias)

    def forward(self, x: Tensor) -> Tensor:
        x = self.act(self.norm(self.fc1(x)))
        x = self.fc2(self.drop(x))

        return x


class _MLPHead(nn.Module):
    def __init__(
        self,
        dim: int,
        num_classes: int = 1000,
        mlp_ratio: int = 3,
        act_layer: Type[nn.Module] = nn.GELU,
        norm_layer: Type[nn.Module] = nn.LayerNorm,
        drop: float = 0.0,
        bias: bool = True,
    ) -> None:
        super().__init__()
        hidden_features = int(mlp_ratio * dim)

        self.fc1 = nn.Linear(dim, hidden_features, bias=bias)
        self.act = act_layer()
        self.norm = norm_layer(hidden_features)

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.drop = nn.Dropout(drop)
        self.fc2 = nn.Linear(hidden_features, num_classes, bias=bias)

    def forward(self, x: Tensor) -> Tensor:
        x = self.avgpool(x)
        x = x.reshape(x.shape[0], -1)

        x = self.norm(self.act(self.fc1(x)))
        x = self.fc2(self.drop(x))

        return x


class _IncepNeXtBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        token_mixer: Type[nn.Module] = nn.Identity,
        norm_layer: Type[nn.Module] = nn.BatchNorm2d,
        mlp_layer: Type[nn.Module] = _ConvMLP,
        mlp_ratio: int = 4,
        act_layer: Type[nn.Module] = nn.GELU,
        ls_init_value: float = 1e-6,
        drop_path: float = 0.0,
    ) -> None:
        super().__init__()
        self.token_mixer = token_mixer(dim)
        self.norm = norm_layer(dim)
        self.mlp = mlp_layer(dim, int(mlp_ratio * dim), act_layer=act_layer)

        self.gamma = (
            nn.Parameter(ls_init_value * lucid.ones(dim)) if ls_init_value else None
        )
        self.drop_path = nn.DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

    def forward(self, x: Tensor) -> Tensor:
        shortcut = x
        x = self.token_mixer(x)
        x = self.norm(x)
        x = self.mlp(x)

        if self.gamma is not None:
            x *= self.gamma.reshape(1, -1, 1, 1)

        x = self.drop_path(x) + shortcut
        return x


class _IncepNeXtStage(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        ds_stride: int = 2,
        depth: int = 2,
        drop_path_rates: list[float] | None = None,
        ls_init_value: float = 0.0,
        token_mixer: Type[nn.Module] = nn.Identity,
        act_layer: Type[nn.Module] = nn.GELU,
        norm_layer: Type[nn.Module] = nn.BatchNorm2d,
        mlp_ratio: int = 4,
    ) -> None:
        super().__init__()
        if ds_stride > 1:
            self.downsample = nn.Sequential(
                norm_layer(in_channels),
                nn.Conv2d(
                    in_channels,
                    out_channels,
                    kernel_size=ds_stride,
                    stride=ds_stride,
                ),
            )
        else:
            self.downsample = nn.Identity()

        drop_path_rates = drop_path_rates or [0.0] * depth
        stage_blocks = []

        for i in range(depth):
            block = _IncepNeXtBlock(
                out_channels,
                token_mixer,
                norm_layer,
                mlp_ratio=mlp_ratio,
                act_layer=act_layer,
                ls_init_value=ls_init_value,
                drop_path=drop_path_rates[i],
            )
            stage_blocks.append(block)
            in_channels = out_channels

        self.blocks = nn.Sequential(*stage_blocks)

    def forward(self, x: Tensor) -> Tensor:
        x = self.downsample(x)
        x = self.blocks(x)
        return x


class InceptionNeXt(nn.Module):
    def __init__(self, config: InceptionNeXtConfig) -> None:
        super().__init__()
        self.config = config
        num_stage = len(config.depths)

        self.num_classes = config.num_classes
        self.drop_rate = config.drop_rate
        self.stem = nn.Sequential(
            nn.Conv2d(3, config.dims[0], kernel_size=4, stride=4),
            nn.BatchNorm2d(config.dims[0]),
        )

        self.stages = None
        dp_rates = [
            x.item()
            for x in lucid.linspace(0, config.drop_path_rate, sum(config.depths))
        ]
        stages = []
        prev_channels = config.dims[0]
        cur = 0

        for i in range(num_stage):
            out_channels = config.dims[i]
            stage = _IncepNeXtStage(
                prev_channels,
                out_channels,
                ds_stride=2 if i > 0 else 1,
                depth=config.depths[i],
                drop_path_rates=dp_rates[cur : cur + config.depths[i]],
                ls_init_value=config.ls_init_value,
                token_mixer=config.token_mixers[i],
                norm_layer=nn.BatchNorm2d,
                mlp_ratio=config.mlp_ratios[i],
            )
            stages.append(stage)
            prev_channels = out_channels
            cur += config.depths[i]

        self.stages = nn.Sequential(*stages)

        self.num_features = prev_channels
        self.head = config.head_fn(
            self.num_features,
            config.num_classes,
            drop=config.drop_rate,
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.stem(x)
        x = self.stages(x)
        x = self.head(x)

        return x


def _raise_for_locked_factory_kwargs(
    kwargs: dict[str, object],
    locked_fields: set[str],
    message: str,
) -> None:
    if locked_fields & kwargs.keys():
        raise TypeError(message)


def _build_inception_next_config(
    *,
    num_classes: int,
    depths: tuple[int, ...],
    dims: tuple[int, ...],
    token_mixers: object,
    **kwargs,
) -> InceptionNeXtConfig:
    return InceptionNeXtConfig(
        num_classes=num_classes,
        depths=depths,
        dims=dims,
        token_mixers=token_mixers,
        **kwargs,
    )


@register_model
def inception_next_atto(num_classes: int = 1000, **kwargs) -> InceptionNeXt:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"depths", "dims", "token_mixers"},
        "factory variants do not allow overriding preset depths, dims, or token_mixers",
    )
    config = _build_inception_next_config(
        num_classes=num_classes,
        depths=(2, 2, 6, 2),
        dims=(40, 80, 160, 320),
        token_mixers=partial(_InceptionDWConv2d, band_kernel_size=9),
        **kwargs,
    )
    return InceptionNeXt(config)


@register_model
def inception_next_tiny(num_classes: int = 1000, **kwargs) -> InceptionNeXt:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"depths", "dims", "token_mixers"},
        "factory variants do not allow overriding preset depths, dims, or token_mixers",
    )
    config = _build_inception_next_config(
        num_classes=num_classes,
        depths=(3, 3, 9, 3),
        dims=(96, 192, 384, 768),
        token_mixers=_InceptionDWConv2d,
        **kwargs,
    )
    return InceptionNeXt(config)


@register_model
def inception_next_small(num_classes: int = 1000, **kwargs) -> InceptionNeXt:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"depths", "dims", "token_mixers"},
        "factory variants do not allow overriding preset depths, dims, or token_mixers",
    )
    config = _build_inception_next_config(
        num_classes=num_classes,
        depths=(3, 3, 27, 3),
        dims=(96, 192, 384, 768),
        token_mixers=_InceptionDWConv2d,
        **kwargs,
    )
    return InceptionNeXt(config)


@register_model
def inception_next_base(num_classes: int = 1000, **kwargs) -> InceptionNeXt:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"depths", "dims", "token_mixers"},
        "factory variants do not allow overriding preset depths, dims, or token_mixers",
    )
    config = _build_inception_next_config(
        num_classes=num_classes,
        depths=(3, 3, 27, 3),
        dims=(128, 256, 512, 1024),
        token_mixers=_InceptionDWConv2d,
        **kwargs,
    )
    return InceptionNeXt(config)



import math
from typing import Callable
from copy import deepcopy

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor

__all__ = [
    "TransformerEncoderLayer",
    "TransformerDecoderLayer",
    "TransformerEncoder",
    "TransformerDecoder",
    "Transformer",
    "SinusoidalPosEmbedding",
    "LearnedPosEmbedding",
    "RotaryPosEmbedding",
]


@nn.auto_repr(
    "d_model",
    "num_heads",
    "dim_feedforward",
    "dropout",
    "activation",
    "layer_norm_eps",
    "norm_first",
    "bias",
)
class TransformerEncoderLayer(nn.Module):
    def __init__(
        self,
        d_model: int,
        num_heads: int,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: Callable[[Tensor], Tensor] = F.relu,
        layer_norm_eps: float = 1e-5,
        norm_first: bool = False,
        bias: bool = True,
    ) -> None:
        super().__init__()
        self.self_attn = nn.MultiHeadAttention(
            d_model, num_heads, dropout=dropout, bias=bias
        )

        self.linear1 = nn.Linear(d_model, dim_feedforward, bias=bias)
        self.linear2 = nn.Linear(dim_feedforward, d_model, bias=bias)

        self.dropout = nn.Dropout(dropout)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)

        self.norm1 = nn.LayerNorm(d_model, eps=layer_norm_eps)
        self.norm2 = nn.LayerNorm(d_model, eps=layer_norm_eps)

        self.activation = activation
        self.norm_first = norm_first

    def _sa_block(
        self,
        x: Tensor,
        src_mask: Tensor | None,
        src_key_padding_mask: Tensor | None,
        is_causal: bool,
        kv_cache: nn.KVCache | None = None,
        use_cache: bool = False,
        cache_position: Tensor | None = None,
        cache_layer_idx: int | None = None,
    ) -> Tensor:
        attn_output = self.self_attn(
            x,
            x,
            x,
            src_key_padding_mask,
            src_mask,
            is_causal,
            kv_cache=kv_cache,
            use_cache=use_cache,
            cache_position=cache_position,
            cache_layer_idx=cache_layer_idx,
        )
        attn_output = self.dropout1(attn_output)

        return attn_output

    def _ff_block(self, x: Tensor) -> Tensor:
        x = self.linear1(x)
        x = self.activation(x)
        x = self.dropout(x)
        x = self.linear2(x)
        x = self.dropout2(x)

        return x

    def forward(
        self,
        src: Tensor,
        src_mask: Tensor | None = None,
        src_key_padding_mask: Tensor | None = None,
        is_causal: bool = False,
        kv_cache: nn.KVCache | None = None,
        use_cache: bool = False,
        cache_position: Tensor | None = None,
        cache_layer_idx: int | None = None,
    ) -> Tensor:
        if self.norm_first:
            x = src + self._sa_block(
                self.norm1(src),
                src_mask,
                src_key_padding_mask,
                is_causal,
                kv_cache=kv_cache,
                use_cache=use_cache,
                cache_position=cache_position,
                cache_layer_idx=cache_layer_idx,
            )
            x += self._ff_block(self.norm2(x))
        else:
            x = self.norm1(
                src
                + self._sa_block(
                    src,
                    src_mask,
                    src_key_padding_mask,
                    is_causal,
                    kv_cache=kv_cache,
                    use_cache=use_cache,
                    cache_position=cache_position,
                    cache_layer_idx=cache_layer_idx,
                )
            )
            x = self.norm2(x + self._ff_block(x))

        return x


@nn.auto_repr(
    "d_model",
    "num_heads",
    "dim_feedforward",
    "dropout",
    "activation",
    "layer_norm_eps",
    "norm_first",
    "bias",
)
class TransformerDecoderLayer(nn.Module):
    def __init__(
        self,
        d_model: int,
        num_heads: int,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: Callable[[Tensor], Tensor] = F.relu,
        layer_norm_eps: float = 1e-5,
        norm_first: bool = False,
        bias: bool = True,
    ) -> None:
        super().__init__()
        self.self_attn = nn.MultiHeadAttention(
            d_model, num_heads, dropout=dropout, bias=bias
        )
        self.multihead_attn = nn.MultiHeadAttention(
            d_model, num_heads, dropout=dropout, bias=bias
        )

        self.linear1 = nn.Linear(d_model, dim_feedforward, bias=bias)
        self.linear2 = nn.Linear(dim_feedforward, d_model, bias=bias)

        self.dropout = nn.Dropout(dropout)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)
        self.dropout3 = nn.Dropout(dropout)

        self.norm1 = nn.LayerNorm(d_model, eps=layer_norm_eps)
        self.norm2 = nn.LayerNorm(d_model, eps=layer_norm_eps)
        self.norm3 = nn.LayerNorm(d_model, eps=layer_norm_eps)

        self.activation = activation
        self.norm_first = norm_first

    def _sa_block(
        self,
        x: Tensor,
        tgt_mask: Tensor | None,
        tgt_key_padding_mask: Tensor | None,
        is_causal: bool,
        kv_cache: nn.KVCache | None = None,
        use_cache: bool = False,
        cache_position: Tensor | None = None,
        cache_layer_idx: int | None = None,
    ) -> Tensor:
        attn_output = self.self_attn(
            x,
            x,
            x,
            tgt_key_padding_mask,
            tgt_mask,
            is_causal,
            kv_cache=kv_cache,
            use_cache=use_cache,
            cache_position=cache_position,
            cache_layer_idx=cache_layer_idx,
        )
        attn_output = self.dropout1(attn_output)

        return attn_output

    def _mha_block(
        self,
        x: Tensor,
        memory: Tensor,
        mem_mask: Tensor | None,
        mem_key_padding_mask: Tensor | None,
        is_causal: bool,
    ) -> Tensor:
        attn_output = self.multihead_attn(
            x, memory, memory, mem_key_padding_mask, mem_mask, is_causal
        )
        attn_output = self.dropout2(attn_output)

        return attn_output

    def _ff_block(self, x: Tensor) -> Tensor:
        x = self.linear1(x)
        x = self.activation(x)
        x = self.dropout(x)
        x = self.linear2(x)
        x = self.dropout3(x)

        return x

    def forward(
        self,
        tgt: Tensor,
        memory: Tensor,
        tgt_mask: Tensor | None = None,
        mem_mask: Tensor | None = None,
        tgt_key_padding_mask: Tensor | None = None,
        mem_key_padding_mask: Tensor | None = None,
        tgt_is_causal: bool = False,
        mem_is_causal: bool = False,
        kv_cache: nn.KVCache | None = None,
        use_cache: bool = False,
        cache_position: Tensor | None = None,
        cache_layer_idx: int | None = None,
    ) -> Tensor:
        if self.norm_first:
            x = tgt + self._sa_block(
                self.norm1(tgt),
                tgt_mask,
                tgt_key_padding_mask,
                tgt_is_causal,
                kv_cache=kv_cache,
                use_cache=use_cache,
                cache_position=cache_position,
                cache_layer_idx=cache_layer_idx,
            )
            x += self._mha_block(
                self.norm2(x), memory, mem_mask, mem_key_padding_mask, mem_is_causal
            )
            x += self._ff_block(self.norm3(x))
        else:
            x = self.norm1(
                tgt
                + self._sa_block(
                    tgt,
                    tgt_mask,
                    tgt_key_padding_mask,
                    tgt_is_causal,
                    kv_cache=kv_cache,
                    use_cache=use_cache,
                    cache_position=cache_position,
                    cache_layer_idx=cache_layer_idx,
                )
            )
            x = self.norm2(
                x
                + self._mha_block(
                    x, memory, mem_mask, mem_key_padding_mask, mem_is_causal
                )
            )
            x = self.norm3(x + self._ff_block(x))

        return x


class TransformerEncoder(nn.Module):
    def __init__(
        self,
        encoder_layer: TransformerEncoderLayer | nn.Module,
        num_layers: int,
        norm: nn.Module | None = None,
    ) -> None:
        super().__init__()
        self.layers = nn.ModuleList(
            [deepcopy(encoder_layer) for _ in range(num_layers)]
        )
        self.num_layers = num_layers
        self.norm = norm

    def forward(
        self,
        src: Tensor,
        src_mask: Tensor | None = None,
        src_key_padding_mask: Tensor | None = None,
        is_causal: bool = False,
        kv_cache: nn.KVCache | None = None,
        use_cache: bool = False,
        cache_position: Tensor | None = None,
        cache_start_layer_idx: int = 0,
    ) -> Tensor:
        output = src
        for i, layer in enumerate(self.layers):
            output = layer(
                output,
                src_mask=src_mask,
                src_key_padding_mask=src_key_padding_mask,
                is_causal=is_causal,
                kv_cache=kv_cache,
                use_cache=use_cache,
                cache_position=cache_position,
                cache_layer_idx=cache_start_layer_idx + i,
            )
        if self.norm is not None:
            output = self.norm(output)

        return output

    def extra_repr(self) -> str:
        return f"num_layers={self.num_layers}, norm={self.norm is not None}"


class TransformerDecoder(nn.Module):
    def __init__(
        self,
        decoder_layer: TransformerDecoderLayer | nn.Module,
        num_layers: int,
        norm: nn.Module | None = None,
    ) -> None:
        super().__init__()
        self.layers = nn.ModuleList(
            [deepcopy(decoder_layer) for _ in range(num_layers)]
        )
        self.num_layers = num_layers
        self.norm = norm

    def forward(
        self,
        tgt: Tensor,
        memory: Tensor,
        tgt_mask: Tensor | None = None,
        mem_mask: Tensor | None = None,
        tgt_key_padding_mask: Tensor | None = None,
        mem_key_padding_mask: Tensor | None = None,
        tgt_is_causal: bool = False,
        mem_is_causal: bool = False,
        kv_cache: nn.KVCache | None = None,
        use_cache: bool = False,
        cache_position: Tensor | None = None,
        cache_start_layer_idx: int = 0,
    ) -> Tensor:
        output = tgt
        for i, layer in enumerate(self.layers):
            output = layer(
                output,
                memory,
                tgt_mask=tgt_mask,
                mem_mask=mem_mask,
                tgt_key_padding_mask=tgt_key_padding_mask,
                mem_key_padding_mask=mem_key_padding_mask,
                tgt_is_causal=tgt_is_causal,
                mem_is_causal=mem_is_causal,
                kv_cache=kv_cache,
                use_cache=use_cache,
                cache_position=cache_position,
                cache_layer_idx=cache_start_layer_idx + i,
            )
        if self.norm is not None:
            output = self.norm(output)

        return output

    def extra_repr(self) -> str:
        return f"num_layers={self.num_layers}, norm={self.norm is not None}"


@nn.auto_repr(
    "d_model",
    "num_heads",
    "num_encoder_layers",
    "num_decoder_layers",
    "dim_feedforward",
    "dropout",
    "activation",
    "layer_norm_eps",
    "norm_first",
    "bias",
)
class Transformer(nn.Module):
    def __init__(
        self,
        d_model: int = 512,
        num_heads: int = 8,
        num_encoder_layers: int = 6,
        num_decoder_layers: int = 6,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: Callable[[Tensor], Tensor] = F.relu,
        layer_norm_eps: float = 1e-5,
        norm_first: bool = False,
        bias: bool = True,
        custom_encoder: nn.Module | None = None,
        custom_decoder: nn.Module | None = None,
    ) -> None:
        super().__init__()
        if custom_encoder is None:
            encoder_layer = TransformerEncoderLayer(
                d_model,
                num_heads,
                dim_feedforward,
                dropout,
                activation,
                layer_norm_eps,
                norm_first,
                bias,
            )
            self.encoder = TransformerEncoder(
                encoder_layer,
                num_encoder_layers,
                norm=nn.LayerNorm(d_model, eps=layer_norm_eps),
            )
        else:
            self.encoder = custom_encoder

        if custom_decoder is None:
            decoder_layer = TransformerDecoderLayer(
                d_model,
                num_heads,
                dim_feedforward,
                dropout,
                activation,
                layer_norm_eps,
                norm_first,
                bias,
            )
            self.decoder = TransformerDecoder(
                decoder_layer,
                num_decoder_layers,
                norm=nn.LayerNorm(d_model, eps=layer_norm_eps),
            )
        else:
            self.decoder = custom_decoder

    def forward(
        self,
        src: Tensor,
        tgt: Tensor,
        src_mask: Tensor | None = None,
        tgt_mask: Tensor | None = None,
        mem_mask: Tensor | None = None,
        src_key_padding_mask: Tensor | None = None,
        tgt_key_padding_mask: Tensor | None = None,
        mem_key_padding_mask: Tensor | None = None,
        kv_cache: nn.KVCache | None = None,
        use_cache: bool = False,
        cache_position: Tensor | None = None,
        cache_start_layer_idx: int = 0,
        encoder_kv_cache: nn.KVCache | None = None,
        use_encoder_cache: bool = False,
        encoder_cache_position: Tensor | None = None,
        encoder_cache_start_layer_idx: int = 0,
    ) -> Tensor:
        memory = self.encoder(
            src,
            src_mask=src_mask,
            src_key_padding_mask=src_key_padding_mask,
            is_causal=False,
            kv_cache=encoder_kv_cache,
            use_cache=use_encoder_cache,
            cache_position=encoder_cache_position,
            cache_start_layer_idx=encoder_cache_start_layer_idx,
        )
        output = self.decoder(
            tgt,
            memory,
            tgt_mask=tgt_mask,
            mem_mask=mem_mask,
            tgt_key_padding_mask=tgt_key_padding_mask,
            mem_key_padding_mask=mem_key_padding_mask,
            tgt_is_causal=False,
            mem_is_causal=False,
            kv_cache=kv_cache,
            use_cache=use_cache,
            cache_position=cache_position,
            cache_start_layer_idx=cache_start_layer_idx,
        )
        return output


class SinusoidalPosEmbedding(nn.Module):
    def __init__(
        self, seq_len: int | None = None, embed_dim: int | None = None
    ) -> None:
        super().__init__()
        self.seq_len = seq_len
        self.embed_dim = embed_dim

    def forward(self, input_: lucid.FloatTensor) -> lucid.FloatTensor:
        if input_.ndim not in {2, 3}:
            raise ValueError(f"{type(self).__name__} expects 2D and 3D Tensor inputs.")

        if input_.dtype.base_dtype is not float:
            raise TypeError(f"'{type(self).__name__}' expects a float Tensor.")

        _is_batched = True if input_.ndim == 3 else False
        x = input_ if _is_batched else input_.unsqueeze(axis=0)

        seq_len = x.shape[1] if self.seq_len is None else self.seq_len
        embed_dim = x.shape[2] if self.embed_dim is None else self.embed_dim

        if seq_len != x.shape[1]:
            raise ValueError(
                f"seq_len does not match: '{self.seq_len}', '{x.shape[1]}'. "
                f"Use 'seq_len=None' for dynamic length inference."
            )
        if embed_dim != x.shape[2]:
            raise ValueError(
                f"embed_dim does not match: '{self.embed_dim}', '{x.shape[2]}'. "
                f"Use 'embed_dim=None' for dynamic length inference."
            )

        posemb = F.sinusoidal_pos_embedding(seq_len, embed_dim, x.device, x.dtype)
        embed = x + posemb
        if not _is_batched:
            embed = embed.squeeze(axis=0)

        return embed


class LearnedPosEmbedding(nn.Module):
    def __init__(self, max_len: int, embed_dim: int) -> None:
        super().__init__()
        self.max_len = max_len
        self.embed_dim = embed_dim
        self.pos_emb = nn.Embedding(max_len, embed_dim)

    def forward(self, input_: lucid.FloatTensor, offset: int = 0) -> lucid.FloatTensor:
        if input_.ndim not in {2, 3}:
            raise ValueError(f"{type(self).__name__} expects 2D and 3D Tensor inputs.")

        if input_.dtype.base_dtype is not float:
            raise TypeError(f"'{type(self).__name__}' expects a float Tensor.")

        seq_len, embed_dim = input_.shape[-2:]
        if embed_dim != self.embed_dim:
            raise ValueError(
                f"embed_dim does not match: '{self.embed_dim}', '{embed_dim}'."
            )

        if offset < 0 or seq_len + offset > self.max_len:
            raise ValueError(f"Invalid position.")

        pos_ids = lucid.arange(
            offset, offset + seq_len, device=input_.device, dtype=lucid.Long
        )
        pos_emb = self.pos_emb(pos_ids)
        if input_.ndim == 3:
            pos_emb = pos_emb.unsqueeze(axis=0)

        return input_ + pos_emb


class RotaryPosEmbedding(nn.Module):
    def __init__(
        self,
        embed_dim: int | None = None,
        max_seq_len: int | None = None,
        interleaved: bool = True,
    ) -> None:
        super().__init__()
        self.embed_dim = embed_dim
        self.max_seq_len = max_seq_len
        self.interleaved = interleaved
        self._cache_seq_len = 0

        self.cos_cached: nn.Buffer | None
        self.sin_cached: nn.Buffer | None
        self.register_buffer("cos_cached", None)
        self.register_buffer("sin_cached", None)

    def _build_cache(self, seq_len: int, embed_dim: int, device: str) -> None:
        theta = lucid.exp(
            -2
            * lucid.arange(embed_dim // 2, device=device, dtype=lucid.Double)
            * (math.log(10000.0) / embed_dim)
        )
        indices = lucid.arange(seq_len, device=device, dtype=lucid.Double)
        freq_half = indices.unsqueeze(-1) @ theta.unsqueeze(0)

        if self.interleaved:
            freq = freq_half.repeat(2, axis=-1)
        else:
            freq = lucid.concatenate([freq_half, freq_half], axis=-1)

        self.cos_cached = nn.Buffer(lucid.cos(freq))
        self.sin_cached = nn.Buffer(lucid.sin(freq))
        self._cache_seq_len = seq_len

    def forward(
        self, input_: lucid.FloatTensor, position_ids: lucid.LongTensor | None = None
    ) -> lucid.FloatTensor:
        if input_.ndim < 2:
            raise ValueError(f"{type(self).__name__} expects input ndim >= 2.")

        if input_.dtype.base_dtype is not float:
            raise TypeError(f"'{type(self).__name__}' expects a float Tensor.")

        seq_len = input_.shape[-2]
        embed_dim = input_.shape[-1]

        if self.embed_dim is not None:
            if embed_dim != self.embed_dim:
                raise ValueError(
                    f"Expected embed_dim of '{self.embed_dim}', got '{embed_dim}'"
                )

        if embed_dim % 2 != 0:
            raise ValueError(
                f"Expected even input embedding dimension, got '{embed_dim}'."
            )

        if position_ids is not None:
            if position_ids.ndim != 1 or position_ids.shape[0] != seq_len:
                raise ValueError(
                    "position_ids must be 1-D with length equal to input_.shape[-2]."
                )
            max_pos = int(lucid.max(position_ids).item())
            if max_pos < 0:
                raise ValueError("position_ids must be non-negative.")
            required_len = max_pos + 1
        else:
            required_len = seq_len

        if self.max_seq_len is not None and required_len > self.max_seq_len:
            raise ValueError(
                f"Required sequence length '{required_len}' exceeds max_seq_len "
                f"'{self.max_seq_len}'."
            )

        if (
            self.cos_cached is None
            or self.sin_cached is None
            or self._cache_seq_len < required_len
            or self.cos_cached.shape[-1] != embed_dim
            or self.cos_cached.device != input_.device
        ):
            build_len = required_len if self.max_seq_len is None else self.max_seq_len
            self._build_cache(build_len, embed_dim, input_.device)

        if position_ids is None:
            cos = self.cos_cached[:seq_len]
            sin = self.sin_cached[:seq_len]
        else:
            ids = position_ids.to(input_.device).astype(lucid.Long)
            cos = self.cos_cached[ids]
            sin = self.sin_cached[ids]

        x = input_.astype(lucid.Double)
        cos = cos.astype(lucid.Double)
        sin = sin.astype(lucid.Double)

        x_rot = lucid.zeros_like(x)
        if self.interleaved:
            x_rot[..., 0::2] = -x[..., 1::2]
            x_rot[..., 1::2] = x[..., 0::2]
        else:
            half = embed_dim // 2
            x_rot[..., :half] = -x[..., half:]
            x_rot[..., half:] = x[..., :half]

        out = x * cos + x_rot * sin
        return out.astype(input_.dtype)



import json
import setuptools
from setuptools import Extension
from pathlib import Path

try:
    import pybind11
except ImportError as e:
    raise RuntimeError(
        "pybind11 is required to build C++ extensions. "
        "Install it first: pip install pybind11"
    ) from e


with open("README.md", "r") as fh:
    long_description = fh.read()

_raw_version = "2.15.5"
PACKAGE_VERSION = "0.0.0" if _raw_version.startswith("{{") else _raw_version

BUILD_CFG_PATH = Path(".extensions/cpp_extensions.json")
with BUILD_CFG_PATH.open("r", encoding="utf-8") as f:
    build_cfg = json.load(f)

common = build_cfg.get("common", {})
extensions_cfg = build_cfg.get("extensions", [])
if not isinstance(extensions_cfg, list) or not extensions_cfg:
    raise RuntimeError(f"Invalid extension config: {BUILD_CFG_PATH}")

ext_modules: list[Extension] = []
for ext_cfg in extensions_cfg:
    ext_modules.append(
        Extension(
            ext_cfg["name"],
            sources=ext_cfg["sources"],
            include_dirs=[pybind11.get_include()],
            language=common.get("language", "c++"),
            extra_compile_args=common.get("extra_compile_args", ["-std=c++20"]),
        )
    )

setuptools.setup(
    name="lucid-dl",
    version=PACKAGE_VERSION,
    license="MIT",
    author="ChanLumerico",
    author_email="greensox284@gmail.com",
    description="Lumerico's Comprehensive Interface for Deep Learning",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ChanLumerico/lucid",
    packages=setuptools.find_namespace_packages(include=["lucid*"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
        "Operating System :: OS Independent",
        "Operating System :: MacOS :: MacOS X",
    ],
    python_requires=">=3.12",
    install_requires=[
        "numpy",
        "pandas",
        "openml",
        "mlx; platform_system == 'Darwin' and platform_machine == 'arm64'",
    ],
    package_data={
        "lucid.weights": ["registry.json", "__init__.pyi"],
        "lucid.models": ["registry.json"],
    },
    ext_modules=ext_modules,
    include_package_data=True,
)



from dataclasses import dataclass
from typing import ClassVar, Type, override

import lucid
import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor

__all__ = [
    "ConvNeXt",
    "ConvNeXtConfig",
    "ConvNeXt_V2",
    "ConvNeXtV2Config",
    "convnext_tiny",
    "convnext_small",
    "convnext_base",
    "convnext_large",
    "convnext_xlarge",
    "convnext_v2_atto",
    "convnext_v2_femto",
    "convnext_v2_pico",
    "convnext_v2_nano",
    "convnext_v2_tiny",
    "convnext_v2_base",
    "convnext_v2_large",
    "convnext_v2_huge",
]


def _normalize_stage_spec(
    values: tuple[int, ...] | list[int],
    name: str,
) -> tuple[int, int, int, int]:
    normalized = tuple(values)
    if len(normalized) != 4:
        raise ValueError(f"{name} must contain exactly 4 stage values")
    if any(not isinstance(value, int) or value <= 0 for value in normalized):
        raise ValueError(f"{name} values must be positive integers")
    return normalized  # type: ignore[return-value]


@dataclass
class ConvNeXtConfig:
    num_classes: int = 1000
    depths: tuple[int, int, int, int] | list[int] = (3, 3, 9, 3)
    dims: tuple[int, int, int, int] | list[int] = (96, 192, 384, 768)
    drop_path: float = 0.0
    layer_scale_init: float = 1e-6

    def __post_init__(self) -> None:
        self.depths = _normalize_stage_spec(self.depths, "depths")
        self.dims = _normalize_stage_spec(self.dims, "dims")
        if self.num_classes <= 0:
            raise ValueError("num_classes must be greater than 0")
        if self.drop_path < 0 or self.drop_path > 1:
            raise ValueError("drop_path must be in the range [0, 1]")
        if self.layer_scale_init < 0:
            raise ValueError("layer_scale_init must be greater than or equal to 0")


@dataclass
class ConvNeXtV2Config:
    num_classes: int = 1000
    depths: tuple[int, int, int, int] | list[int] = (3, 3, 9, 3)
    dims: tuple[int, int, int, int] | list[int] = (96, 192, 384, 768)
    drop_path: float = 0.0

    def __post_init__(self) -> None:
        self.depths = _normalize_stage_spec(self.depths, "depths")
        self.dims = _normalize_stage_spec(self.dims, "dims")
        if self.num_classes <= 0:
            raise ValueError("num_classes must be greater than 0")
        if self.drop_path < 0 or self.drop_path > 1:
            raise ValueError("drop_path must be in the range [0, 1]")


class _Block(nn.Module):
    def __init__(
        self, in_channels: int, drop_path: float = 0.0, layer_scale_init: float = 1e-6
    ) -> None:
        super().__init__()

        self.dwconv = nn.Conv2d(
            in_channels, in_channels, kernel_size=7, padding=3, groups=in_channels
        )
        self.norm = nn.LayerNorm(in_channels, eps=1e-6)

        self.pwconv1 = nn.Linear(in_channels, 4 * in_channels)
        self.gelu = nn.GELU()
        self.pwconv2 = nn.Linear(4 * in_channels, in_channels)

        self.gamma = (
            nn.Parameter(layer_scale_init * lucid.ones(in_channels))
            if layer_scale_init > 0
            else None
        )
        self.drop_path = nn.DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

    def forward(self, x: Tensor) -> Tensor:
        input_ = x
        x = self.dwconv(x)
        x = x.transpose((0, 2, 3, 1))
        x = self.norm(x)

        n, h, w, _ = x.shape
        x = self.pwconv1(x.reshape(n * h * w, -1))
        x = self.gelu(x)
        x = self.pwconv2(x)

        if self.gamma is not None:
            flat_gamma = lucid.repeat(self.gamma, x.shape[0], axis=0)
            x = flat_gamma * x.reshape(-1)

        x = x.reshape(n, -1, h, w)
        x = input_ + self.drop_path(x)
        return x


class _ChannelsFisrtLayerNorm(nn.LayerNorm):
    def __init__(
        self,
        normalized_shape: int,
        eps: float = 0.00001,
        elementwise_affine: bool = True,
        bias: bool = True,
    ) -> None:
        super().__init__(normalized_shape, eps, elementwise_affine, bias)

    @override
    def forward(self, x: Tensor) -> Tensor:
        x = x.transpose((0, 2, 3, 1))
        out = super().forward(x)
        out = out.transpose((0, 3, 1, 2))
        return out


class ConvNeXt(nn.Module):
    base_block: ClassVar[Type[nn.Module]] = _Block

    def __init__(self, config: ConvNeXtConfig) -> None:
        super().__init__()
        self.config = config

        self.downsample_layers = nn.ModuleList()
        stem = nn.Sequential(
            nn.Conv2d(3, config.dims[0], kernel_size=4, stride=4),
            _ChannelsFisrtLayerNorm(config.dims[0]),
        )
        self.downsample_layers.append(stem)

        for i in range(3):
            downsample = nn.Sequential(
                _ChannelsFisrtLayerNorm(config.dims[i]),
                nn.Conv2d(config.dims[i], config.dims[i + 1], kernel_size=2, stride=2),
            )
            self.downsample_layers.append(downsample)

        self.stages = nn.ModuleList()
        dp_rates = [
            x.item() for x in lucid.linspace(0, config.drop_path, sum(config.depths))
        ]
        cur = 0
        for i in range(4):
            stage = nn.Sequential(
                *[
                    self.base_block(
                        config.dims[i],
                        dp_rates[cur + j],
                        config.layer_scale_init,
                    )
                    for j in range(config.depths[i])
                ]
            )
            self.stages.append(stage)
            cur += config.depths[i]

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.norm = nn.LayerNorm(config.dims[-1])
        self.head = nn.Linear(config.dims[-1], config.num_classes)

    def forward(self, x: Tensor) -> Tensor:
        for i in range(4):
            x = self.downsample_layers[i](x)
            x = self.stages[i](x)

        x = self.avgpool(x)
        x = x.reshape(x.shape[0], -1)

        x = self.norm(x)
        x = self.head(x)

        return x


class _Block_V2(nn.Module):
    def __init__(self, channels: int, drop_path: float = 0.0, *args) -> None:
        super().__init__()
        self.dwconv = nn.Conv2d(
            channels, channels, kernel_size=7, padding=3, groups=channels
        )
        self.norm = nn.LayerNorm(channels)

        self.pwconv1 = nn.Linear(channels, 4 * channels)
        self.gelu = nn.GELU()
        self.grn = nn.GlobalResponseNorm(4 * channels)

        self.pwconv2 = nn.Linear(4 * channels, channels)
        self.drop_path = nn.DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

    def forward(self, x: Tensor) -> Tensor:
        input_ = x
        x = self.dwconv(x).transpose((0, 2, 3, 1))
        x = self.norm(x)

        n, h, w, _ = x.shape
        x = self.pwconv1(x.reshape(n * h * w, -1))
        x = self.gelu(x)
        x = self.grn(x.reshape(n, -1, h, w))
        x = self.pwconv2(x.reshape(n * h * w, -1))

        x = x.reshape(n, -1, h, w)
        x = input_ + self.drop_path(x)

        return x


class ConvNeXt_V2(ConvNeXt):
    base_block: ClassVar[Type[nn.Module]] = _Block_V2

    def __init__(self, config: ConvNeXtV2Config) -> None:
        super().__init__(
            ConvNeXtConfig(
                num_classes=config.num_classes,
                depths=config.depths,
                dims=config.dims,
                drop_path=config.drop_path,
            )
        )
        self.config = config


def _raise_for_locked_factory_kwargs(
    kwargs: dict[str, object],
    locked_fields: set[str],
    message: str,
) -> None:
    if locked_fields & kwargs.keys():
        raise TypeError(message)


def _build_convnext_config(
    *,
    num_classes: int,
    depths: tuple[int, int, int, int],
    dims: tuple[int, int, int, int],
    **kwargs,
) -> ConvNeXtConfig:
    return ConvNeXtConfig(
        num_classes=num_classes,
        depths=depths,
        dims=dims,
        **kwargs,
    )


def _build_convnext_v2_config(
    *,
    num_classes: int,
    depths: tuple[int, int, int, int],
    dims: tuple[int, int, int, int],
    **kwargs,
) -> ConvNeXtV2Config:
    return ConvNeXtV2Config(
        num_classes=num_classes,
        depths=depths,
        dims=dims,
        **kwargs,
    )


@register_model
def convnext_tiny(num_classes: int = 1000, **kwargs) -> ConvNeXt:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"depths", "dims"},
        "factory variants do not allow overriding preset depths or dims",
    )
    config = _build_convnext_config(
        num_classes=num_classes,
        depths=(3, 3, 9, 3),
        dims=(96, 192, 384, 768),
        **kwargs,
    )
    return ConvNeXt(config)


@register_model
def convnext_small(num_classes: int = 1000, **kwargs) -> ConvNeXt:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"depths", "dims"},
        "factory variants do not allow overriding preset depths or dims",
    )
    config = _build_convnext_config(
        num_classes=num_classes,
        depths=(3, 3, 27, 3),
        dims=(96, 192, 364, 768),
        **kwargs,
    )
    return ConvNeXt(config)


@register_model
def convnext_base(num_classes: int = 1000, **kwargs) -> ConvNeXt:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"depths", "dims"},
        "factory variants do not allow overriding preset depths or dims",
    )
    config = _build_convnext_config(
        num_classes=num_classes,
        depths=(3, 3, 27, 3),
        dims=(128, 256, 512, 1024),
        **kwargs,
    )
    return ConvNeXt(config)


@register_model
def convnext_large(num_classes: int = 1000, **kwargs) -> ConvNeXt:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"depths", "dims"},
        "factory variants do not allow overriding preset depths or dims",
    )
    config = _build_convnext_config(
        num_classes=num_classes,
        depths=(3, 3, 27, 3),
        dims=(192, 384, 768, 1536),
        **kwargs,
    )
    return ConvNeXt(config)


@register_model
def convnext_xlarge(num_classes: int = 1000, **kwargs) -> ConvNeXt:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"depths", "dims"},
        "factory variants do not allow overriding preset depths or dims",
    )
    config = _build_convnext_config(
        num_classes=num_classes,
        depths=(3, 3, 27, 3),
        dims=(256, 512, 1024, 2048),
        **kwargs,
    )
    return ConvNeXt(config)


@register_model
def convnext_v2_atto(num_classes: int = 1000, **kwargs) -> ConvNeXt_V2:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"depths", "dims"},
        "factory variants do not allow overriding preset depths or dims",
    )
    config = _build_convnext_v2_config(
        num_classes=num_classes,
        depths=(2, 2, 6, 2),
        dims=(40, 80, 160, 320),
        **kwargs,
    )
    return ConvNeXt_V2(config)


@register_model
def convnext_v2_femto(num_classes: int = 1000, **kwargs) -> ConvNeXt_V2:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"depths", "dims"},
        "factory variants do not allow overriding preset depths or dims",
    )
    config = _build_convnext_v2_config(
        num_classes=num_classes,
        depths=(2, 2, 6, 2),
        dims=(48, 96, 192, 384),
        **kwargs,
    )
    return ConvNeXt_V2(config)


@register_model
def convnext_v2_pico(num_classes: int = 1000, **kwargs) -> ConvNeXt_V2:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"depths", "dims"},
        "factory variants do not allow overriding preset depths or dims",
    )
    config = _build_convnext_v2_config(
        num_classes=num_classes,
        depths=(2, 2, 6, 2),
        dims=(64, 128, 256, 512),
        **kwargs,
    )
    return ConvNeXt_V2(config)


@register_model
def convnext_v2_nano(num_classes: int = 1000, **kwargs) -> ConvNeXt_V2:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"depths", "dims"},
        "factory variants do not allow overriding preset depths or dims",
    )
    config = _build_convnext_v2_config(
        num_classes=num_classes,
        depths=(2, 2, 8, 2),
        dims=(80, 160, 320, 640),
        **kwargs,
    )
    return ConvNeXt_V2(config)


@register_model
def convnext_v2_tiny(num_classes: int = 1000, **kwargs) -> ConvNeXt_V2:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"depths", "dims"},
        "factory variants do not allow overriding preset depths or dims",
    )
    config = _build_convnext_v2_config(
        num_classes=num_classes,
        depths=(3, 3, 9, 3),
        dims=(96, 192, 384, 768),
        **kwargs,
    )
    return ConvNeXt_V2(config)


@register_model
def convnext_v2_base(num_classes: int = 1000, **kwargs) -> ConvNeXt_V2:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"depths", "dims"},
        "factory variants do not allow overriding preset depths or dims",
    )
    config = _build_convnext_v2_config(
        num_classes=num_classes,
        depths=(3, 3, 27, 3),
        dims=(128, 256, 512, 1024),
        **kwargs,
    )
    return ConvNeXt_V2(config)


@register_model
def convnext_v2_large(num_classes: int = 1000, **kwargs) -> ConvNeXt_V2:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"depths", "dims"},
        "factory variants do not allow overriding preset depths or dims",
    )
    config = _build_convnext_v2_config(
        num_classes=num_classes,
        depths=(3, 3, 27, 3),
        dims=(192, 384, 768, 1536),
        **kwargs,
    )
    return ConvNeXt_V2(config)


@register_model
def convnext_v2_huge(num_classes: int = 1000, **kwargs) -> ConvNeXt_V2:
    _raise_for_locked_factory_kwargs(
        kwargs,
        {"depths", "dims"},
        "factory variants do not allow overriding preset depths or dims",
    )
    config = _build_convnext_v2_config(
        num_classes=num_classes,
        depths=(3, 3, 27, 3),
        dims=(352, 704, 1408, 2816),
        **kwargs,
    )
    return ConvNeXt_V2(config)



import inspect
import threading
from typing import Any

from lucid._jit.ir import IRGraph, IRNode, IRValue, ShapeSpec

_trace_local = threading.local()


def is_tracing() -> bool:
    return getattr(_trace_local, "active", False)


def _get_active_tracer() -> TracingContext | None:
    return getattr(_trace_local, "tracer", None)


def _capture_op_init_state(op_self: Any) -> tuple[tuple, dict]:
    base_attrs = {"result", "_inplace", "_inplace_target", "_flops"}
    try:
        sig = inspect.signature(type(op_self).__init__)
        params = list(sig.parameters.values())[1:]

        positional_vals = []
        for p in params:
            if p.name in base_attrs:
                continue
            if hasattr(op_self, p.name):
                positional_vals.append(getattr(op_self, p.name))

        return tuple(positional_vals), {}

    except (ValueError, TypeError):
        return (), {}


class TracingContext:
    def __init__(self) -> None:
        self._value_counter: int = 0
        self._tensor_to_vid: dict[int, int] = {}

        self.values: dict[int, IRValue] = {}
        self.nodes: list[IRNode] = []

        self._node_counter: int = 0
        self.input_ids: list[int] = []
        self.param_ids: set[int] = set()

    def register_tensor(
        self, tensor: Any, *, is_input: bool = False, is_param: bool = False
    ) -> int:
        tid = id(tensor)
        if tid in self._tensor_to_vid:
            return self._tensor_to_vid[tid]

        vid = self._value_counter
        self._value_counter += 1

        spec = ShapeSpec(
            shape=tuple(tensor.shape),
            dtype=tensor.dtype,
            device=tensor.device,
        )
        self.values[vid] = IRValue(value_id=vid, shape_spec=spec, live_tensor=tensor)
        self._tensor_to_vid[tid] = vid

        if is_input:
            self.input_ids.append(vid)
        if is_param:
            self.param_ids.add(vid)

        return vid

    def record_op(
        self,
        op_self: Any,
        input_tensors: tuple,
        output_tensors: tuple,
        non_tensor_args: tuple,
        non_tensor_kwargs: dict,
        device: str,
        has_gradient: bool,
        n_in: int | None,
        n_ret: int | None,
    ) -> None:
        input_ids = [self.register_tensor(t) for t in input_tensors]

        output_ids = []
        for t in output_tensors:
            tid = id(t)
            if tid in self._tensor_to_vid:
                output_ids.append(self._tensor_to_vid[tid])
            else:
                vid = self._value_counter
                self._value_counter += 1
                spec = ShapeSpec(
                    shape=tuple(t.shape),
                    dtype=t.dtype,
                    device=t.device,
                )
                self.values[vid] = IRValue(value_id=vid, shape_spec=spec, live_tensor=t)
                self._tensor_to_vid[tid] = vid
                output_ids.append(vid)

        op_init_args, op_init_kwargs = _capture_op_init_state(op_self)

        node = IRNode(
            node_id=self._node_counter,
            op_class=type(op_self),
            op_init_args=op_init_args,
            op_init_kwargs=op_init_kwargs,
            input_ids=input_ids,
            output_ids=output_ids,
            non_tensor_args=non_tensor_args,
            non_tensor_kwargs=non_tensor_kwargs,
            device=device,
            has_gradient=has_gradient,
            n_in=n_in,
            n_ret=n_ret,
        )
        self.nodes.append(node)
        self._node_counter += 1

    def finalize(self, output_tensors: tuple) -> IRGraph:
        output_ids = []
        for t in output_tensors:
            tid = id(t)
            if tid in self._tensor_to_vid:
                output_ids.append(self._tensor_to_vid[tid])

        return IRGraph(
            input_ids=list(self.input_ids),
            output_ids=output_ids,
            values=self.values,
            nodes=self.nodes,
            param_ids=self.param_ids,
        )



import importlib
from collections.abc import Sequence
from typing import Any, Callable
from pathlib import Path
import sys

__all__ = ["run_tests"]


def _require_module(module_name: str, message: str) -> None:
    try:
        importlib.import_module(module_name)
    except ModuleNotFoundError as exc:
        raise ModuleNotFoundError(message) from exc


def _get_pytest_main() -> Callable[..., int]:
    return importlib.import_module("pytest").main


def run_tests(
    paths: str | Path | Sequence[str | Path] = "lucid/test",
    *,
    raise_on_fail: bool = False,
    show_progress: bool = True,
    concise_failures: bool = True,
    single_line: bool = True,
) -> int:
    _require_module("torch", "'torch' is required to perform pytest on Lucid.")
    _require_module("pytest", "'pytest' is required to run tests.")
    test_paths = _normalize_test_paths(paths)
    return _run_pytest_with_progress(
        test_paths,
        show_progress=show_progress,
        raise_on_fail=raise_on_fail,
        concise_failures=concise_failures,
        single_line=single_line,
    )


def _normalize_test_paths(paths: str | Path | Sequence[str | Path]) -> list[str]:
    if isinstance(paths, (str, Path)):
        return [str(paths)]
    return [str(p) for p in paths]


def _run_pytest_with_progress(
    paths: list[str],
    *,
    show_progress: bool,
    raise_on_fail: bool,
    concise_failures: bool,
    single_line: bool,
) -> int:
    _ = single_line
    if not show_progress:
        return _run_pytest(paths, raise_on_fail=raise_on_fail)

    try:
        from tqdm import tqdm
    except Exception as exc:
        raise RuntimeError("tqdm is required for progress display.") from exc

    plugin = _TqdmProgressPlugin(tqdm, concise_failures=concise_failures)

    exit_code = _get_pytest_main()(
        ["-q", "-p", "no:terminalreporter", *paths], plugins=[plugin]
    )
    if raise_on_fail and exit_code != 0:
        raise AssertionError(f"pytest exited with code: {exit_code}")
    return exit_code


def _run_pytest(paths: list[str], *, raise_on_fail: bool = False) -> int:
    exit_code = _get_pytest_main()(["-q", *paths])
    if raise_on_fail and exit_code != 0:
        raise AssertionError(f"pytest exited with code: {exit_code}")
    return exit_code


class _PytestProgressPluginBase:
    def __init__(self, concise_failures: bool = True) -> None:
        self._concise_failures: bool = concise_failures
        self._failure_infos: list[tuple[str, str]] = []
        self._passed: int = 0
        self._failed: int = 0
        self._counted: set[str] = set()
        self._total: int = 0
        self._label_chars: int = 42

    @staticmethod
    def _colorize(text: str, color_code: str = "") -> str:
        if not sys.stderr.isatty():
            return text
        return f"\033[{color_code}m{text}\033[0m"

    def _green(self, text: str) -> str:
        return self._colorize(text, "92")

    def _red(self, text: str) -> str:
        return self._colorize(text, "91")

    def _yellow(self, text: str) -> str:
        return self._colorize(text, "93")

    def _cyan(self, text: str) -> str:
        return self._colorize(text, "96")

    def _bold(self, text: str) -> str:
        return self._colorize(text, "1")

    def _get_short_label(self, nodeid: str | None) -> str:
        label = nodeid or "processing"
        max_label_len = max(8, self._label_chars)
        short_label = label.split("::")[-1]
        if len(short_label) > max_label_len:
            short_label = f"{short_label[: max_label_len - 3]}..."
        return short_label

    def _record_case_result(self, report: Any) -> None:
        if report.passed:
            self._passed += 1
        if report.failed:
            self._failed += 1

    def _build_summary_postfix(self, nodeid: str | None) -> str:
        return f"{self._get_short_label(nodeid)} | passed={self._passed} failed={self._failed}"

    def _add_failure(self, report: Any) -> None:
        if not self._concise_failures:
            return
        if report.outcome not in {"failed", "error"}:
            return
        raw = str(report.longrepr).splitlines()
        message = raw[-1] if raw else "failed"
        self._failure_infos.append((report.nodeid, message))

    def pytest_sessionfinish(self, session: Any, exitstatus: int) -> None:
        print(file=sys.stderr)
        if self._failed == 0:
            summary = self._green("ALL PASSED")
        else:
            summary = self._red(f"FAILED ({self._failed})")
        print(
            self._bold(
                f"{summary} | total={self._total}, passed={self._passed}, failed={self._failed}"
            ),
            file=sys.stderr,
        )

        if self._concise_failures and self._failure_infos:
            print(file=sys.stderr)
            print(self._yellow("Failing cases:"), file=sys.stderr)
            for idx, (nodeid, message) in enumerate(self._failure_infos, start=1):
                print(f" {idx}. {self._cyan(nodeid)}", file=sys.stderr)
                print(f"    {message}", file=sys.stderr)


class _TqdmProgressPlugin(_PytestProgressPluginBase):
    def __init__(self, tqdm_mod: Any, concise_failures: bool = True) -> None:
        super().__init__(concise_failures=concise_failures)
        self._tqdm_mod = tqdm_mod
        self._bar: Any | None = None

    def pytest_collection_modifyitems(
        self, session: Any, config: Any, items: Sequence[Any]
    ) -> None:
        self._total = len(items)
        self._bar = self._tqdm_mod(
            total=len(items),
            desc="Running tests",
            file=sys.stderr,
            unit="case",
            leave=False,
        )

    def _set_postfix(self, nodeid: str | None = None) -> None:
        if self._bar is None:
            return
        self._bar.set_postfix_str(self._build_summary_postfix(nodeid), refresh=False)

    def pytest_runtest_logreport(self, report: Any) -> None:
        if self._bar is None:
            return
        if report.when != "call":
            return
        if report.nodeid in self._counted:
            return

        self._counted.add(report.nodeid)
        self._record_case_result(report)
        self._set_postfix(report.nodeid)
        self._add_failure(report)
        self._bar.update(1)

    def pytest_sessionfinish(self, session: Any, exitstatus: int) -> None:
        if self._bar is not None:
            self._bar.close()
        super().pytest_sessionfinish(session=session, exitstatus=exitstatus)



import math

from dataclasses import dataclass, replace
from typing import Literal

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor

__all__ = ["UNetConfig", "UNetStageConfig", "UNet", "ResUNet"]


@dataclass
class UNetStageConfig:
    channels: int
    num_blocks: int = 2
    kernel_size: int = 3
    dilation: int = 1
    use_attention: bool = False
    dropout: float = 0.0

    def __post_init__(self) -> None:
        if self.channels <= 0:
            raise ValueError("channels must be greater than 0.")
        if self.num_blocks <= 0:
            raise ValueError("num_blocks must be greater than 0.")
        if self.kernel_size <= 0 or self.kernel_size % 2 == 0:
            raise ValueError("kernel_size must be a positive odd integer.")
        if self.dilation <= 0:
            raise ValueError("dilation must be greater than 0.")
        if not 0.0 <= self.dropout < 1.0:
            raise ValueError("dropout must be in [0, 1).")


_UNetNormName = Literal["batch", "group", "instance", "none"]
_UNetActName = Literal["relu", "leaky_relu", "gelu", "silu"]

_UNetBlockName = Literal["basic", "res", "convnext"]
_UNetSkipMerge = Literal["concat", "add"]

_UNetDownsample = Literal["conv", "maxpool", "avgpool"]
_UNetUpsample = Literal["transpose", "bilinear", "nearest"]


@dataclass
class UNetConfig:
    in_channels: int
    out_channels: int

    encoder_stages: tuple[UNetStageConfig, ...] | list[UNetStageConfig]
    decoder_stages: tuple[UNetStageConfig, ...] | list[UNetStageConfig] | None = None
    bottleneck: UNetStageConfig | None = None

    block: _UNetBlockName = "basic"
    norm: _UNetNormName = "batch"
    act: _UNetActName = "relu"

    skip_merge: _UNetSkipMerge = "concat"
    downsample_mode: _UNetDownsample = "conv"
    upsample_mode: _UNetUpsample = "transpose"

    stem_channels: int | None = None
    final_kernel_size: int = 1

    deep_supervision: bool = False
    align_corners: bool = False
    bias: bool | None = None

    def __post_init__(self) -> None:
        self.encoder_stages = tuple(self.encoder_stages)
        if len(self.encoder_stages) == 0:
            raise ValueError("encoder_stages must contain at least one stage.")

        if self.decoder_stages is None:
            self.decoder_stages = tuple(
                UNetStageConfig(
                    channels=stage.channels,
                    num_blocks=stage.num_blocks,
                    kernel_size=stage.kernel_size,
                    dilation=stage.dilation,
                    use_attention=stage.use_attention,
                    dropout=stage.dropout,
                )
                for stage in reversed(self.encoder_stages[:-1])
            )
        else:
            self.decoder_stages = tuple(self.decoder_stages)

        if self.bottleneck is None:
            self.bottleneck = UNetStageConfig(
                channels=self.encoder_stages[-1].channels,
                num_blocks=2,
            )

        if self.in_channels <= 0:
            raise ValueError("in_channels must be greater than 0.")
        if self.out_channels <= 0:
            raise ValueError("out_channels must be greater than 0.")
        if self.stem_channels is not None and self.stem_channels <= 0:
            raise ValueError("stem_channels must be greater than 0.")
        if self.final_kernel_size <= 0 or self.final_kernel_size % 2 == 0:
            raise ValueError("final_kernel_size must be a positive odd integer.")

        expected_decoder_depth = max(len(self.encoder_stages) - 1, 0)
        if len(self.decoder_stages) != expected_decoder_depth:
            raise ValueError(
                "decoder_stages must have 'len(encoder_stages) - 1' stages."
            )

        if self.skip_merge == "add":
            for enc_stage, dec_stage in zip(
                reversed(self.encoder_stages[:-1]), self.decoder_stages
            ):
                if enc_stage.channels != dec_stage.channels:
                    raise ValueError(
                        "skip_merge='add' requires matching encoder/decoder channels."
                    )

        if self.bias is None:
            self.bias = self.norm == "none"

    @classmethod
    def from_channels(
        cls,
        in_channels: int,
        out_channels: int,
        channels: tuple[int, ...] | list[int],
        num_blocks: int | tuple[int, ...] | list[int] = 2,
        **kwargs,
    ) -> UNetConfig:
        channels = tuple(channels)
        if isinstance(num_blocks, int):
            num_blocks = (num_blocks,) * len(channels)
        else:
            num_blocks = tuple(num_blocks)

        if len(channels) == 0 or len(channels) != len(num_blocks):
            raise ValueError(
                "channels and num_blocks must have the same non-zero length."
            )

        encoder_stages = tuple(
            UNetStageConfig(channels=ch, num_blocks=nb)
            for ch, nb in zip(channels, num_blocks)
        )
        return cls(
            in_channels=in_channels,
            out_channels=out_channels,
            encoder_stages=encoder_stages,
            **kwargs,
        )


def _same_padding(kernel_size: int, dilation: int = 1) -> int:
    return dilation * (kernel_size - 1) // 2


def _make_norm(norm: _UNetNormName, channels: int) -> nn.Module:
    if norm == "batch":
        return nn.BatchNorm2d(channels)
    if norm == "instance":
        return nn.InstanceNorm2d(channels)
    if norm == "group":
        num_groups = min(32, channels)
        while channels % num_groups != 0:
            num_groups -= 1
        return nn.GroupNorm(num_groups, channels)
    if norm == "none":
        return nn.Identity()

    raise ValueError(f"Unsupported norm: '{norm}'")


def _make_act(act: _UNetActName) -> nn.Module:
    if act == "relu":
        return nn.ReLU()
    if act == "leaky_relu":
        return nn.LeakyReLU()
    if act == "gelu":
        return nn.GELU()
    if act == "silu":
        return nn.Swish()
    raise ValueError(f"Unsupported activation: '{act}'")


class _BasicBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        *,
        kernel_size: int,
        dilation: int,
        norm: _UNetNormName,
        act: _UNetActName,
        dropout: float,
        bias: bool,
    ) -> None:
        super().__init__()
        padding = _same_padding(kernel_size, dilation)
        self.block = nn.Sequential(
            nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size=kernel_size,
                padding=padding,
                dilation=dilation,
                bias=bias,
            ),
            _make_norm(norm, out_channels),
            _make_act(act),
            nn.Dropout2d(dropout) if dropout > 0.0 else nn.Identity(),
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.block(x)


class _ResBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        *,
        kernel_size: int,
        dilation: int,
        norm: _UNetNormName,
        act: _UNetActName,
        dropout: float,
        bias: bool,
    ) -> None:
        super().__init__()
        padding = _same_padding(kernel_size, dilation)

        self.norm1 = _make_norm(norm, in_channels)
        self.act1 = _make_act(act)
        self.conv1 = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            padding=padding,
            dilation=dilation,
            bias=bias,
        )

        self.norm2 = _make_norm(norm, out_channels)
        self.act2 = _make_act(act)
        self.dropout = nn.Dropout2d(dropout) if dropout > 0.0 else nn.Identity()
        self.conv2 = nn.Conv2d(
            out_channels,
            out_channels,
            kernel_size=kernel_size,
            padding=padding,
            dilation=dilation,
            bias=bias,
        )

        self.shortcut = (
            nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=bias)
            if in_channels != out_channels
            else nn.Identity()
        )

    def forward(self, x: Tensor) -> Tensor:
        h = self.conv1(self.act1(self.norm1(x)))
        h = self.conv2(self.dropout(self.act2(self.norm2(h))))
        return h + self.shortcut(x)


class _SelfAttention2d(nn.Module):
    def __init__(self, channels: int, *, norm: _UNetNormName, bias: bool) -> None:
        super().__init__()
        self.norm = _make_norm(norm, channels)
        self.qkv = nn.Conv2d(channels, channels * 3, kernel_size=1, bias=bias)
        self.proj = nn.Conv2d(channels, channels, kernel_size=1, bias=bias)

    def forward(self, x: Tensor) -> Tensor:
        b, c, h, w = x.shape
        q, k, v = self.qkv(self.norm(x)).chunk(3, axis=1)

        q = q.reshape(b, c, h * w)
        k = k.reshape(b, c, h * w)
        v = v.reshape(b, c, h * w)

        attn = (q.mT @ k) / math.sqrt(c)
        attn = F.softmax(attn, axis=-1)

        out = v @ attn.mT
        out = out.reshape(b, c, h, w)

        return x + self.proj(out)


class _UNetStage(nn.Module):
    def __init__(
        self,
        in_channels: int,
        stage_config: UNetStageConfig,
        *,
        block: _UNetBlockName,
        norm: _UNetNormName,
        act: _UNetActName,
        bias: bool,
    ) -> None:
        super().__init__()
        if block == "basic":
            block_cls = _BasicBlock
        elif block == "res":
            block_cls = _ResBlock
        else:
            raise ValueError(
                f"Unsupported block type: {block}.",
                "This implementation supports only 'basic' and 'res'.",
            )

        layers = []
        cur_channels = in_channels
        for _ in range(stage_config.num_blocks):
            layers.append(
                block_cls(
                    cur_channels,
                    stage_config.channels,
                    kernel_size=stage_config.kernel_size,
                    dilation=stage_config.dilation,
                    norm=norm,
                    act=act,
                    dropout=stage_config.dropout,
                    bias=bias,
                )
            )
            cur_channels = stage_config.channels

        self.blocks = nn.Sequential(*layers)
        self.attn = (
            _SelfAttention2d(stage_config.channels, norm=norm, bias=bias)
            if stage_config.use_attention
            else nn.Identity()
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.blocks(x)
        x = self.attn(x)
        return x


class _Downsample(nn.Module):
    def __init__(self, channels: int, mode: _UNetDownsample, *, bias: bool) -> None:
        super().__init__()
        if mode == "conv":
            self.op = nn.Conv2d(
                channels, channels, kernel_size=3, stride=2, padding=1, bias=bias
            )
        elif mode == "maxpool":
            self.op = nn.MaxPool2d(kernel_size=2, stride=2)
        elif mode == "avgpool":
            self.op = nn.AvgPool2d(kernel_size=2, stride=2)
        else:
            raise ValueError(f"Unsupported downsample mode: '{mode}'")

    def forward(self, x: Tensor) -> Tensor:
        return self.op(x)


class _Upsample(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        mode: _UNetUpsample,
        *,
        align_corners: bool,
        bias: bool,
    ) -> None:
        super().__init__()
        self.mode = mode
        self.align_corners = align_corners

        if mode == "transpose":
            self.op = nn.ConvTranspose2d(
                in_channels, out_channels, kernel_size=2, stride=2, bias=bias
            )
        elif mode in {"bilinear", "nearest"}:
            self.op = nn.Sequential(
                nn.Upsample(
                    scale_factor=2,
                    mode=mode,
                    align_corners=align_corners if mode == "bilinear" else False,
                ),
                nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=bias),
            )
        else:
            raise ValueError(f"Unsupported upsample mode: '{mode}'")

    def forward(self, x: Tensor, target_size: tuple[int, int] | None = None) -> Tensor:
        x = self.op(x)
        if target_size is not None and target_size != x.shape[-2:]:
            x = F.interpolate(
                x, size=target_size, mode="bilinear", align_corners=self.align_corners
            )
        return x


class UNet(nn.Module):
    def __init__(self, config: UNetConfig) -> None:
        super().__init__()
        self.config = config

        stem_channels = (
            config.stem_channels
            if config.stem_channels is not None
            else config.encoder_stages[0].channels
        )
        head_padding = _same_padding(config.final_kernel_size)

        self.stem = nn.Sequential(
            nn.Conv2d(
                config.in_channels,
                stem_channels,
                kernel_size=3,
                padding=1,
                bias=config.bias,
            ),
            _make_norm(config.norm, stem_channels),
            _make_act(config.act),
        )
        self.encoder = nn.ModuleList()
        self.downsamplers = nn.ModuleList()

        cur_channels = stem_channels
        for idx, stage_cfg in enumerate(config.encoder_stages):
            self.encoder.append(
                _UNetStage(
                    cur_channels,
                    stage_cfg,
                    block=config.block,
                    norm=config.norm,
                    act=config.act,
                    bias=config.bias,
                )
            )
            cur_channels = stage_cfg.channels

            if idx < len(config.encoder_stages) - 1:
                self.downsamplers.append(
                    _Downsample(cur_channels, config.downsample_mode, bias=config.bias)
                )

        self.bottleneck = _UNetStage(
            cur_channels,
            config.bottleneck,
            block=config.block,
            norm=config.norm,
            act=config.act,
            bias=config.bias,
        )
        cur_channels = config.bottleneck.channels

        self.upsamplers = nn.ModuleList()
        self.decoder = nn.ModuleList()

        skip_channels = [stage.channels for stage in config.encoder_stages[:-1]]
        reversed_skip_channels = list(reversed(skip_channels))

        for dec_cfg, skip_ch in zip(config.decoder_stages, reversed_skip_channels):
            self.upsamplers.append(
                _Upsample(
                    cur_channels,
                    dec_cfg.channels,
                    config.upsample_mode,
                    align_corners=config.align_corners,
                    bias=config.bias,
                )
            )

            stage_in_channels = (
                dec_cfg.channels + skip_ch
                if config.skip_merge == "concat"
                else dec_cfg.channels
            )

            self.decoder.append(
                _UNetStage(
                    stage_in_channels,
                    dec_cfg,
                    block=config.block,
                    norm=config.norm,
                    act=config.act,
                    bias=config.bias,
                )
            )
            cur_channels = dec_cfg.channels

        self.head = nn.Conv2d(
            cur_channels,
            config.out_channels,
            kernel_size=config.final_kernel_size,
            padding=head_padding,
            bias=True,
        )

        self.aux_heads = nn.ModuleList()
        if config.deep_supervision:
            for dec_cfg in config.decoder_stages[:-1]:
                self.aux_heads.append(
                    nn.Conv2d(
                        dec_cfg.channels, config.out_channels, kernel_size=1, bias=True
                    )
                )

    def _merge_skip(self, x: Tensor, skip: Tensor) -> Tensor:
        if self.config.skip_merge == "concat":
            return lucid.concatenate([x, skip], axis=1)
        if self.config.skip_merge == "add":
            return x + skip
        raise ValueError(f"Unsupported skip_merge: '{self.config.skip_merge}'")

    def _resize_to_input(self, x: Tensor, input_size: tuple[int, int]) -> Tensor:
        if x.shape[-2:] == input_size:
            return x
        return F.interpolate(
            x, size=input_size, mode="bilinear", align_corners=self.config.align_corners
        )

    def forward(self, x: Tensor) -> Tensor | dict[str, Tensor | list[Tensor]]:
        input_size = x.shape[-2:]
        x = self.stem(x)

        skips: list[Tensor] = []
        for idx, stage in enumerate(self.encoder):
            x = stage(x)
            skips.append(x)

            if idx < len(self.downsamplers):
                x = self.downsamplers[idx](x)

        x = self.bottleneck(x)

        aux_outputs: list[Tensor] = []
        decode_skips = list(reversed(skips[:-1]))

        for idx, (upsample, stage, skip) in enumerate(
            zip(self.upsamplers, self.decoder, decode_skips)
        ):
            x = upsample(x, target_size=skip.shape[-2:])
            x = self._merge_skip(x, skip)
            x = stage(x)

            if self.config.deep_supervision and idx < len(self.aux_heads):
                aux = self.aux_heads[idx](x)
                aux_outputs.append(self._resize_to_input(aux, input_size))

        out = self.head(x)
        out = self._resize_to_input(out, input_size)

        if self.config.deep_supervision:
            return {"out": out, "aux": aux_outputs}

        return out


class ResUNet(UNet):
    def __init__(self, config: UNetConfig) -> None:
        if config.block != "res":
            config = replace(config, block="res")
        super().__init__(config)
